import click
from omibio.cli.plot_cli import plot_group
from omibio.viz import plot_kmer
from omibio.bio import KmerResult
import matplotlib.pyplot as plt
from typing import TextIO
from csv import DictReader
import os


@plot_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None,
    help="Directory to save plots"
)
@click.option(
    "--no-show",
    is_flag=True,
    help="Whether not to show the plots."
)
def kmer(
    source: TextIO,
    output: str,
    no_show: bool,
):
    """Plot k-mer counts from a TSV file."""

    if output is not None:
        os.makedirs(output, exist_ok=True)

    results: dict[str, KmerResult] = {}
    reader = DictReader(source, delimiter="\t")
    for row in reader:
        seq_id = str(row["seq_id"])
        if seq_id not in results:
            results[seq_id] = KmerResult(
                k=int(row["k"]), counts={row["kmer"]: int(row["count"])},
                seq_id=seq_id
            )
        else:
            if int(row["k"]) != results[seq_id].k:
                raise ValueError(
                    "Got inconsistent k values: "
                    f"{row["k"]} vs {results[seq_id].k}"
                )
            results[seq_id].counts[row["kmer"]] = int(row["count"])

    plot_kmer(list(results.values()))

    if output is not None:
        plt.tight_layout()
        output_path = os.path.join(
            output, "kmer_heatmap.png"
        )
        plt.savefig(output_path)
        if no_show:
            plt.close()

    if not no_show:
        plt.tight_layout()
        plt.show()



import click


@click.group()
def fastq_group():
    """FASTQ related tools."""
    pass


def register_commands():
    from .fastq_to_fasta import to_fasta
    from .fastq_view import view
    from .fastq_info import info

    fastq_group.add_command(to_fasta)
    fastq_group.add_command(view)
    fastq_group.add_command(info)


register_commands()



from omibio.sequence.seq_utils.clean import clean, CleanReport, CleanReportItem
from omibio.sequence.seq_utils.complement import complement, reverse_complement
from omibio.sequence.seq_utils.random_seq import random_seq, random_fasta
from omibio.sequence.seq_utils.shuffle_seq import shuffle_seq
from omibio.sequence.seq_utils.transcribe import transcribe, reverse_transcribe
from omibio.sequence.seq_utils.translate import translate_nt


__all__ = [
    "clean",
    "CleanReport",
    "CleanReportItem",
    "complement",
    "reverse_complement",
    "random_seq",
    "random_fasta",
    "shuffle_seq",
    "transcribe",
    "reverse_transcribe",
    "translate_nt"
]



import click
from omibio.cli.orf_cli import orf_group
from omibio.io import read_fasta
from omibio.analysis.find_orfs import find_orfs
from omibio.sequence import Polypeptide
from typing import TextIO
import csv


@orf_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "--min-length", "-min",
    type=int,
    default=100,
    help="Minimum length of ORFs to consider. Defaults to 0."
)
@click.option(
    "--max-length", "-max",
    type=int,
    default=10000,
    help="Maximum length of ORFs to consider. Defaults to 10000."
)
@click.option(
    "--start-codons",
    type=str,
    default="ATG",
    help="Comma-separated start codons (e.g., ATG,GTG). Default: ATG."
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None,
    help="Output file path"
)
@click.option(
    "--overlap",
    is_flag=True,
    help="Whether to allow overlapping ORFs."
)
@click.option(
    "--no-reverse",
    is_flag=True,
    help="Whether to include reverse strand ORFs."
)
@click.option(
    "--no-sort",
    is_flag=True,
    help="Whether to sort the results by length in descending order."
)
@click.option(
    "--translate", "-t",
    is_flag=True,
    help=(
        "Whether to translate nucleotide sequences to amino acid sequences."
        " (shown only if --show-seq is used)."
    )
)
@click.option(
    "--show-seq",
    is_flag=True,
    help="Whether to show orf sequence."
)
def find(
    source: TextIO,
    output: str | None,
    min_length: int,
    max_length: int,
    overlap: bool,
    no_reverse: bool,
    no_sort: bool,
    translate: bool,
    start_codons: str,
    show_seq: bool
) -> None:
    """Find orfs of sequences from a FASTA file."""

    start_codon_set = {
        codon.strip().upper() for codon in start_codons.split(",")
    }

    seqs = read_fasta(source).seq_dict()
    all_orfs = []

    for seq_id, seq_obj in seqs.items():
        if isinstance(seq_obj, Polypeptide):
            raise TypeError(
                "Cannot find ORFs in amino acid sequences"
            )
        orfs = find_orfs(
            seq=seq_obj,
            min_length=min_length,
            max_length=max_length,
            overlap=overlap,
            include_reverse=not no_reverse,
            sort_by_length=not no_sort,
            translate=translate,
            start_codons=start_codon_set,
            seq_id=seq_id
        )
        all_orfs.extend(orfs.intervals)

    if not no_sort:
        all_orfs.sort(key=lambda x: x.length, reverse=True)

    res: list[list[str | None]] = []

    for orf in all_orfs:
        frame = f"{orf.frame:+}" if orf.frame > 0 else orf.frame
        base_fields = [
            orf.seq_id,
            str(orf.start),
            str(orf.end),
            orf.strand,
            str(frame),
            str(orf.length)
        ]
        if show_seq:
            nt_seq = str(orf.nt_seq) if orf.nt_seq is not None else "None"
            aa_seq = str(orf.aa_seq) if orf.aa_seq is not None else "None"
            base_fields.extend([nt_seq, aa_seq])
        res.append(base_fields)

    header = ["seq_id", "start", "end", "strand", "frame", "length"]
    if show_seq:
        header.append("nt_seq")
        if translate:
            header.append("aa_seq")

    rows = [header] + res

    if output is None:
        for row in rows:
            click.echo("\t".join(str(f) for f in row))
    else:
        with open(output, "w", newline="", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerows(rows)
        click.echo(f"Result written to {output}")



from omibio.sequence.polypeptide import Polypeptide


class Sequence:
    """
    A class representing a DNA or RNA sequence with methods for analysis.
    """

    _VALID_DNA_BASES = {
        "A", "T", "C", "G",
        "N", "R", "Y", "K", "M", "B", "V", "D", "H", "S", "W"
    }
    _VALID_RNA_BASES = {
        "A", "U", "C", "G",
        "N", "R", "Y", "K", "M", "B", "V", "D", "H", "S", "W"
    }

    def __init__(
        self,
        sequence: str | None = None,
        rna: bool | None = None,
        strict: bool = False
    ):

        if rna is not None and not isinstance(rna, bool):
            raise TypeError(
                "Sequence argument 'rna' must be bool or None, got "
                + type(rna).__name__
            )
        if not isinstance(strict, bool):
            raise TypeError(
                "Sequence argument 'strict' must be bool, got "
                + type(strict).__name__
            )

        self._is_rna = rna
        self._strict = strict
        self.sequence = sequence if sequence is not None else ""

    @property
    def strict(self) -> bool:
        """Getter, returns whether strict mode is enabled."""
        return self._strict

    @property
    def is_rna(self) -> bool | None:
        """Getter, returns whether the sequence is RNA (True), DNA (False),
        or unknown (None).
        """
        return self._is_rna

    @property
    def type(self) -> str:
        """Return 'DNA' or 'RNA' indicating the sequence type."""
        return "RNA" if self._is_rna else "DNA"

    @property
    def sequence(self) -> str:
        """Getter, returns the sequence."""
        return self._sequence

    @sequence.setter
    def sequence(self, sequence: str) -> None:
        """Setter, based on the strict rule, performs a sequence base validity
        check and automatically determines the DNA/RNA type.

        Args:
            sequence (str): The nucleotide sequence.

        Raises:
            TypeError:
                if sequence is not str
            ValueError:
                If strict mode is enabled and invalid bases are found
            ValueError:
                If strict mode is enabled and both 'T' and 'U' are found
        """
        if sequence is None:
            sequence = ""
        elif not isinstance(sequence, str):
            raise TypeError(
                "Sequence argument 'sequence' must be str, not "
                + type(sequence).__name__
            )

        sequence = sequence.upper()  # Convert to uppercase

        if self._strict:
            if self._is_rna is None:
                contains_t = "T" in sequence
                contains_u = "U" in sequence
                if contains_t and contains_u:
                    raise ValueError(
                        "(Strict Mode) "
                        "Ambiguous sequence: contains both 'T' and 'U'"
                    )
                valid_bases = (
                    self._VALID_RNA_BASES if contains_u
                    else self._VALID_DNA_BASES
                )
            else:
                valid_bases = (
                    self._VALID_RNA_BASES if self._is_rna
                    else self._VALID_DNA_BASES
                )

            if invalid := set(sequence) - valid_bases:
                seq_typ = (
                    "RNA" if valid_bases == self._VALID_RNA_BASES else "DNA"
                )
                # Validate sequence contains only A, C, G, T，N
                raise ValueError(
                    f"(Strict Mode) Invalid base(s) for {seq_typ} "
                    f"sequence found: {invalid}"
                )

        self._sequence = sequence

        if self._is_rna is None:
            self._is_rna = "U" in sequence

    def gc_content(self, percent: bool = False) -> float | str:
        """Calculate and return the GC content of the sequence."""
        seq_length = len(self.sequence)
        if seq_length == 0:
            return 0.0 if not percent else "0.00%"

        gc = self.sequence.count("G") + self.sequence.count("C")

        return (round(gc / seq_length, 4) if not percent
                else f"{(gc / seq_length) * 100:.2f}%")

    def at_content(self, percent: bool = False) -> float | str:
        """Calculate and return the AT content of the sequence."""
        seq_length = len(self.sequence)
        if seq_length == 0:
            return 0.0 if not percent else "0.00%"

        at = self.sequence.count("A") + self.sequence.count("T")

        return (round(at / seq_length, 4) if not percent
                else f"{(at / seq_length) * 100:.2f}%")

    def complement(self) -> "Sequence":
        """Return the complement of the sequence."""
        if self._is_rna is True:
            comp_table = str.maketrans("AUCGRYKMBVDHSWN", "UAGCYRMKVBHDSWN")
        else:
            comp_table = str.maketrans("ATCGRYKMBVDHSWN", "TAGCYRMKVBHDSWN")
        comp = self.sequence.translate(comp_table)
        return Sequence(comp, rna=self._is_rna, strict=self._strict)

    def reverse_complement(self) -> "Sequence":
        """Return the reverse complement of the sequence."""
        if self._is_rna is True:
            rev_comp_tb = str.maketrans("AUCGRYKMBVDHSWN", "UAGCYRMKVBHDSWN")
        else:
            rev_comp_tb = str.maketrans("ATCGRYKMBVDHSWN", "TAGCYRMKVBHDSWN")
        rev_comp = self.sequence.translate(rev_comp_tb)[::-1]
        return Sequence(rev_comp, rna=self._is_rna, strict=self._strict)

    def transcribe(self, strand: str = "+") -> "Sequence":
        """Transcribe the DNA sequence to RNA."""
        if self._is_rna is True:
            return self
        if strand not in {"+", "-"}:
            raise ValueError("strand type should be either '+' or '-'")
        if strand == "+":
            rna_seq = self.sequence.replace("T", "U")
        else:
            rna_seq = self.reverse_complement().sequence.replace("T", "U")
        return Sequence(rna_seq, rna=True, strict=self._strict)

    def reverse_transcribe(self) -> "Sequence":
        """Transcribe the RNA sequence to DNA."""
        if self._is_rna is False:
            return self
        return Sequence(
            self.sequence.replace("U", "T"), rna=False, strict=self._strict
        )

    def subseq(self, start: int, end: int | None = None) -> "Sequence":
        """Return a subsequence from start to end (end exclusive)."""
        if (
            not isinstance(start, int)
            or (end is not None and not isinstance(end, int))
        ):
            raise TypeError("subseq() argument 'start' and ''end' must be int")
        return Sequence(
            self.sequence[start: end], rna=self._is_rna, strict=self._strict
        )

    def translate_nt(
        self,
        strict: bool = False,
        as_str: bool = False,
        stop_symbol: bool = False,
        to_stop: bool = False,
        frame: int = 0,
        require_start: bool = False
    ) -> Polypeptide | str:
        """Translate the nucleotide sequence to an amino acid sequence."""
        from omibio.sequence.seq_utils.translate import translate_nt

        return translate_nt(
            self.sequence,
            as_str=as_str,
            strict=strict,
            stop_symbol=stop_symbol,
            to_stop=to_stop,
            frame=frame,
            require_start=require_start
        )

    def count(self, base: str) -> int:
        """Count occurrences of a base in the sequence."""
        return self.sequence.count(base)

    def copy(
        self,
        as_rna: bool | None = None,
        strict: bool | None = None
    ) -> "Sequence":
        """Return a copy of the Sequence object."""

        new_strict = strict if strict is not None else self.strict
        return Sequence(str(self), rna=as_rna, strict=new_strict)

    def to_strict(self, as_rna: bool | None = None) -> "Sequence":
        """Return a copy of the Sequence object in strict mode."""
        return self.copy(as_rna=as_rna, strict=True)

    def is_valid(self) -> bool:
        """Check if the sequence contains only valid bases."""
        valid_bases = (
            self._VALID_RNA_BASES if self._is_rna
            else self._VALID_DNA_BASES
        )
        return not (set(self.sequence) - valid_bases)

    def __len__(self) -> int:
        """Return the length of the sequence."""
        return len(self.sequence)

    def __str__(self) -> str:
        """Return string representation of the sequence."""
        return self.sequence

    def __repr__(self) -> str:
        from omibio.utils.truncate_repr import truncate_repr
        seq_repr = truncate_repr(self.sequence)
        return (
            f"Sequence({seq_repr}, "
            f"type={self.type}, "
            f"strict={self._strict})"
        )

    def __getitem__(self, idx) -> str:
        return self.sequence[idx]

    def __iter__(self):
        return iter(self.sequence)

    def __contains__(self, item) -> bool:
        return item in self.sequence

    def __eq__(self, other) -> bool:
        if isinstance(other, Sequence):
            if self._strict or other.strict:
                return (
                    self.sequence == other.sequence
                    and self._is_rna == other.is_rna
                )
            else:
                return self.sequence == other.sequence

        elif isinstance(other, str):
            return self.sequence == other.upper()
        return False

    def __add__(self, other) -> "Sequence":
        if isinstance(other, Sequence):
            strict_mode = self._strict or other.strict
            if strict_mode and self._is_rna != other.is_rna:
                raise TypeError(
                    "(Strict Mode) "
                    "Cannot combine RNA sequence and DNA sequence"
                )
            rna_result = self._is_rna
            return Sequence(
                self.sequence + other.sequence,
                rna=rna_result, strict=strict_mode
            )

        elif isinstance(other, str):
            other = other.upper()
            if self._strict:
                if "U" in other and "T" in other:
                    raise ValueError(
                        "(Strict Mode) Invalid string added to Sequence: "
                        "contains both 'U' and 'T'"
                    )
                if (
                    (("U" in other) and not self._is_rna)
                    or (("T" in other) and self._is_rna)
                ):
                    raise TypeError(
                        "(Strict Mode) "
                        "Cannot combine RNA sequence and DNA sequence"
                    )
            return Sequence(
                self.sequence + other, rna=self._is_rna, strict=self._strict
            )

        else:
            raise TypeError(
                "Can only add Sequence or str to Sequence, got "
                + type(other).__name__
            )

    def __mul__(self, n: int) -> "Sequence":
        if not isinstance(n, int):
            raise TypeError(
                f"Can only multiply by int, got {type(n).__name__}"
            )
        if n < 0:
            raise ValueError(
                "Sequence cannot be multiply by a negative number"
            )

        return Sequence(
            self.sequence * n, rna=self._is_rna, strict=self._strict
        )


def main():
    dna = Sequence("AAAATGCATGCTGACTGTAGCTGATTTATTGCTATC")
    print(repr(dna))


if __name__ == "__main__":
    main()



import click
from omibio.cli.fastq_cli import fastq_group
from omibio.io import read_fastq_iter, write_fasta
from typing import TextIO


@fastq_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None
)
@click.option(
    "--line-len", "-l",
    type=int,
    default=60
)
@click.option(
    "--prefix", "-p",
    type=str,
    default=None
)
def to_fasta(
    source: TextIO,
    output: str | None,
    line_len: int,
    prefix: str
):
    """Convert FASTQ to FASTA format."""

    result = read_fastq_iter(source)

    if prefix is not None:
        count = 1
        seqs = {}
        for e in result:
            seqs[f"{prefix}_{count}"] = e.seq
            count += 1
    else:
        seqs = {e.seq_id: e.seq for e in result}

    if output is not None:
        write_fasta(seqs=seqs, file_name=output, line_len=int(line_len))
        click.echo(f"Written to {output}")
    else:
        for line in write_fasta(seqs=seqs, line_len=int(line_len)):
            click.echo(line)



from omibio.sequence.sequence import Sequence
from omibio.sequence.polypeptide import Polypeptide
from omibio.bio.seq_interval import SeqInterval


DNA_CODON_TABLE = {
    'TTT': 'F', 'TTC': 'F',  # Phenylalanine
    'TTA': 'L', 'TTG': 'L',  # Leucine
    'TCT': 'S', 'TCC': 'S', 'TCA': 'S', 'TCG': 'S',  # Serine
    'TAT': 'Y', 'TAC': 'Y',  # Tyrosine
    'TAA': '*', 'TAG': '*',  # Stop codons
    'TGT': 'C', 'TGC': 'C',  # Cysteine
    'TGA': '*',              # Stop codon
    'TGG': 'W',              # Tryptophan

    'CTT': 'L', 'CTC': 'L', 'CTA': 'L', 'CTG': 'L',  # Leucine
    'CCT': 'P', 'CCC': 'P', 'CCA': 'P', 'CCG': 'P',  # Proline
    'CAT': 'H', 'CAC': 'H',  # Histidine
    'CAA': 'Q', 'CAG': 'Q',  # Glutamine
    'CGT': 'R', 'CGC': 'R', 'CGA': 'R', 'CGG': 'R',  # Arginine

    'ATT': 'I', 'ATC': 'I', 'ATA': 'I',  # Isoleucine
    'ATG': 'M',                          # Methionine (Start)
    'ACT': 'T', 'ACC': 'T', 'ACA': 'T', 'ACG': 'T',  # Threonine
    'AAT': 'N', 'AAC': 'N',  # Asparagine
    'AAA': 'K', 'AAG': 'K',  # Lysine
    'AGT': 'S', 'AGC': 'S',  # Serine
    'AGA': 'R', 'AGG': 'R',  # Arginine

    'GTT': 'V', 'GTC': 'V', 'GTA': 'V', 'GTG': 'V',  # Valine
    'GCT': 'A', 'GCC': 'A', 'GCA': 'A', 'GCG': 'A',  # Alanine
    'GAT': 'D', 'GAC': 'D',  # Aspartic acid
    'GAA': 'E', 'GAG': 'E',  # Glutamic acid
    'GGT': 'G', 'GGC': 'G', 'GGA': 'G', 'GGG': 'G',  # Glycine
}


def translate_nt(
    seq: Sequence | str | SeqInterval | None,
    strict: bool = False,
    as_str: bool = False,
    stop_symbol: bool = False,
    to_stop: bool = False,
    frame: int = 0,
    require_start: bool = False
) -> Polypeptide | str:
    """Translate a nucleotide sequence to an amino acid sequence.

    Args:
        seq (Sequence | str):
            Input nucleotide sequence.
        stop_symbol (bool, optional):
            Whether to include stop codon symbol '*'. Defaults to True.
        to_stop (bool, optional):
            Whether to stop translation at the first stop codon.
            Defaults to False.
        frame (int, optional):
            Frame offset (0, 1, or 2). Defaults to 0.
        require_start (bool, optional):
            Whether to start translation at the first start codon.
            Defaults to False.

    Raises:
        ValueError: If frame is not 0, 1, or 2.

    Returns:
        str: Translated amino acid sequence.
    """

    if not isinstance(seq, (Sequence, str, SeqInterval)):
        raise TypeError(
            "find_otranslaterfs() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )

    if not seq or len(seq) < 3:
        return ""

    if frame not in (0, 1, 2):
        raise ValueError(f"Invalid frame: {frame}, Frame must be 0, 1, or 2.")

    aa_seq = []
    start_idx = 0
    seq = str(seq).replace("U", "T")
    seq = seq[frame:]

    if require_start:
        for j in range(0, len(seq) - 2, 3):
            if seq[j: j+3] == "ATG":
                start_idx = j
                break
        else:
            return ""

    for i in range(start_idx, len(seq) - 2, 3):
        codon = seq[i: i + 3]
        amino = DNA_CODON_TABLE.get(codon, "X")

        if amino == "*":
            if stop_symbol:
                aa_seq.append("*")
            if to_stop:
                break
            continue

        aa_seq.append(amino)

    res = "".join(aa_seq)

    return res if as_str else Polypeptide(res, strict=strict)


def main() -> None:
    seq = Sequence("ATGGCCATTGTAATGGGCCGCTGAAAGGGTGACCGATAG")
    prot = translate_nt(seq, strict=False, stop_symbol=True)
    print(prot)


if __name__ == "__main__":
    main()



from omibio.sequence.sequence import Sequence, Polypeptide
from omibio.bio import KmerResult
from collections import Counter
from omibio.sequence.seq_utils.complement import reverse_complement
from omibio.viz import plot_kmer


def kmer(
    seq: Sequence | str | Polypeptide,
    k: int,
    seq_id: str | None = None,
    canonical: bool = False,
    min_count: int = 1,
) -> KmerResult:
    """Count k-mers in a given sequence.

    Args:
        seq (Sequence | str | Polypeptide):
            Input sequence to analyze.
        k (int):
            Length of the k-mers to count.
        seq_id (str | None, optional):
            An optional identifier for the sequence. Defaults to None.
        canonical (bool, optional):
            Whether to count canonical k-mers (considering reverse
            complements). Defaults to False.
        min_count (int, optional):
            Minimum count threshold for k-mers to include in the result.
            Defaults to 1.

    Raises:
        TypeError:
            If the input sequence is not of type Sequence, Polypeptide,
            or string.
        TypeError:
            If k is not of type int.
        TypeError:
            If min_count is not of type int.
        ValueError:
            If k is not a positive integer.
        ValueError:
            If min_count is negative.

    Returns:
        KmerResult:
            An object containing k-mer counts and metadata.
    """

    if canonical:
        cache: dict[str, str] = {}

        def get_canonical(kmer_seq):
            if kmer_seq in cache:
                return cache[kmer_seq]
            rc = reverse_complement(kmer_seq, as_str=True)
            canon = min(kmer_seq, rc)
            cache[kmer_seq] = cache[rc] = canon
            return canon
    else:
        def get_canonical(kmer_seq):
            return kmer_seq

    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "kmer() argument 'seq' must be Sequence or str, got "
            + type(seq).__name__
        )
    if not isinstance(k, int):
        raise TypeError(
            f"kmer() argument 'k' must be int, got {type(k).__name__}"
        )
    if not isinstance(min_count, int):
        raise TypeError(
            "kmer() argument 'min_count' must be int, got "
            + type(min_count).__name__
        )
    if k <= 0:
        raise ValueError(
            f"kmer() argument 'k' must be a positive number, got {k}"
        )
    if min_count < 0:
        raise ValueError(
            "kmer() argument 'min_count' must be a non-negative number, got "
            + str(min_count)
        )

    seq_str = str(seq).upper()
    n = len(seq_str)

    if k > n:
        return KmerResult(
            k=k, counts=dict(), seq_id=seq_id, type="kmer", plot_func=plot_kmer
        )

    kmer_counter: dict = Counter()

    for i in range(n - k + 1):
        curr_kmer = seq_str[i: i+k]
        kmer_counter[get_canonical(curr_kmer)] += 1

    if min_count > 1:
        kmer_counter = Counter(
            {kmer: c for kmer, c in kmer_counter.items() if c >= min_count}
        )
    return KmerResult(
        k=k, counts=kmer_counter, seq_id=seq_id,
        type="kmer", plot_func=plot_kmer
    )


def main():
    from omibio.io import read_fasta
    seq = read_fasta(
        r"./examples/data/example_single_long_seq.fasta"
    )["example"]
    res = kmer(seq, 3, min_count=30, seq_id="test")
    print(res)
    print(repr(res))
    print(res["ACT"])
    res.plot(show=True)


if __name__ == "__main__":
    main()



from omibio.analysis.gc_content import gc
from omibio.analysis.sliding_gc import sliding_gc
from omibio.analysis.find_orfs import find_orfs
from omibio.analysis.consensus import find_consensus
from omibio.analysis.find_motif import find_motifs
from omibio.analysis.kmer import kmer
from omibio.analysis.protein_mass import calc_mass
from omibio.analysis.palindrome import find_palindrome

__all__ = [
    "gc",
    "sliding_gc",
    "find_orfs",
    "find_consensus",
    "find_motifs",
    "kmer",
    "calc_mass",
    "find_palindrome"
]



import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from omibio.bio import SeqInterval
from omibio.bio import IntervalResult


def plot_sliding_gc(
    gc_list: list[SeqInterval] | IntervalResult,
    window_avg: bool = True,
    ax: Axes | None = None,
    show: bool = False,
    seq_id: str | None = None,
    figsize: tuple = (9, 3)
) -> Axes:
    """Plot sliding window GC content along a sequence.

    Args:
        gc_list (list[SeqInterval] | IntervalResult):
            List of GC content intervals or an IntervalResult object.
        window_avg (bool, optional):
            Whether to plot the average GC content across all windows.
            Defaults to True.
        ax (Axes | None, optional):
            Matplotlib Axes to plot on. If None, a new figure and axes
            will be created. Defaults to None.
        show (bool, optional):
            Whether to display the plot immediately. Defaults to False.
        seq_id (str | None, optional):
            Sequence identifier for the title. Defaults to None.
        figsize (tuple, optional):
            Figure size if a new figure is created. Defaults to (9, 3).

    Returns:
        Axes:
            The Matplotlib Axes object containing the plot.
    """

    if not gc_list:
        if ax is None:
            ax = plt.subplots(figsize=figsize)[1]
        return ax

    if isinstance(gc_list, IntervalResult):
        seq_id = seq_id or gc_list.seq_id

    positions = [
        (window.start + window.end) / 2
        for window in gc_list if window.gc is not None
    ]
    gc_vals = [
        window.gc * 100 for window in gc_list
        if window.gc is not None
    ]
    window_average = sum(gc_vals) / len(gc_vals)

    if ax is None:
        ax = plt.subplots(figsize=(9, 3))[1]

    if window_avg:
        ax.axhline(
            y=window_average, color="#E14040",
            linestyle='--', label=f'Window average GC%: {window_average:.2f}'
        )

        ax.plot(positions, gc_vals, color="#005AEB", linewidth=1)

    if seq_id is not None:
        ax.set_title(f"Sliding Window GC% of {seq_id}")
    else:
        ax.set_title("Sliding Window GC%")
    ax.set_xlabel("Position in Sequence")
    ax.set_ylabel("GC%")
    ax.set_ylim(0, 100)
    ax.grid(True, linestyle='--', alpha=0.5)
    ax.legend(loc='upper right')

    if show:
        plt.show()
    return ax


def main():
    from omibio.analysis.sliding_gc import sliding_gc
    from omibio.io.read_fasta import read_fasta
    seq = read_fasta(
        "./examples/data/example_single_long_seq.fasta"
    )["example"]
    gc_list = sliding_gc(seq, seq_id="test")
    plot_sliding_gc(gc_list)
    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from omibio.bio.analysis_result import AnalysisResult
from omibio.bio.seq_interval import SeqInterval
from typing import Iterator


@dataclass
class IntervalResult(AnalysisResult):
    """Store data returned by analytical functions for interval types,
    is a subclass of AnalysisResult.

    Args:
        AnalysisResult:
            Base class for analysis results.

    Raises:
        TypeError:
            If the input types are incorrect.
    """

    intervals: list[SeqInterval] = field(default_factory=list)

    def __post_init__(self):
        if not isinstance(self.intervals, list):
            raise TypeError(
                "IntervalResult argument 'intervals' must be list, got "
                + type(self.intervals).__name__
            )

    def __len__(self) -> int:
        return len(self.intervals)

    def __iter__(self) -> Iterator[SeqInterval]:
        return iter(self.intervals)

    def __getitem__(self, idx: int | slice) -> SeqInterval | list[SeqInterval]:
        return self.intervals[idx]

    def __repr__(self) -> str:
        return (
            f"IntervalResult(intervals={self.intervals!r}, "
            f"seq_id={self.seq_id!r}, type={self.type!r})"
        )

    def __str__(self) -> str:
        return str(self.intervals)


def main():
    from omibio.viz import plot_motifs
    intervals = [
        SeqInterval(0, 4, nt_seq="ACTG"), SeqInterval(8, 12, nt_seq="ACTG")
    ]
    result = IntervalResult(
        intervals=intervals, seq_id="test", type="motif", plot_func=plot_motifs
    )
    print(result)
    print(type(result))


if __name__ == "__main__":
    main()



from importlib.metadata import version
from omibio.sequence import Sequence, Polypeptide
from omibio.bio import (
    SeqInterval,
    SeqCollections,
    SeqEntry,
    AnalysisResult,
    IntervalResult,
    KmerResult
)
from omibio.sequence.seq_utils.clean import CleanReport, CleanReportItem
from omibio.io.read_fasta import FastaFormatError
from omibio.io.read_fastq import FastqFormatError

__version__ = version("omibio")

__all__ = [
    "Sequence", "Polypeptide",
    "SeqInterval", "SeqCollections", "SeqEntry",
    "AnalysisResult", "IntervalResult", "KmerResult",
    "CleanReport", "CleanReportItem",
    "FastaFormatError",
    "FastqFormatError"
]



from omibio.bio import SeqInterval, IntervalResult
from omibio.sequence import Sequence
from omibio.sequence.seq_utils.translate import translate_nt
from omibio.viz import plot_orfs

STOP_CODONS = {"TAA", "TAG", "TGA"}


def find_orfs_in_frame(
    seq: str,
    min_length: int,
    max_length: int,
    overlap: bool,
    strand: str,
    frame: int,
    translate: bool,
    start_codons: set[str],
    seq_id: str | None = None
) -> list[SeqInterval]:
    """Internal helper function. Scan a single frame and return raw ORFs."""

    orf_list = []
    active_starts: list[int] = []

    for i in range(frame, len(seq)-2, 3):
        codon = seq[i: i+3]
        if codon in STOP_CODONS:
            end_idx = i + 3

            for start_idx in active_starts:
                orf_length = i + 3 - start_idx

                if not (min_length <= orf_length <= max_length):
                    continue

                nt_seq = seq[start_idx: end_idx]
                aa_seq = (
                    str(translate_nt(nt_seq, stop_symbol=False)) if translate
                    else None
                )
                orf_list.append(
                    SeqInterval(
                        start=start_idx, end=end_idx,
                        nt_seq=nt_seq, type='ORF',
                        strand=strand, frame=frame+1,
                        aa_seq=aa_seq, seq_id=seq_id
                    )
                )
            active_starts = []
        else:
            if codon in start_codons:
                if overlap:
                    active_starts.append(i)
                else:
                    if not active_starts:
                        active_starts.append(i)

    return orf_list


def find_orfs(
    seq: Sequence | str,
    min_length: int = 100,
    max_length: int = 10000,
    overlap: bool = False,
    include_reverse: bool = True,
    sort_by_length: bool = True,
    translate: bool = False,
    start_codons: set[str] | list[str] | tuple[str] = {"ATG"},
    seq_id: str | None = None
) -> IntervalResult:
    """Find ORFs in a given sequence.

    Args:
        seq (Sequence | str):
            Input sequence.
        min_length (int, optional):
            Minimum length of ORFs to consider. Defaults to 100.
        max_length (int, optional):
            Maximum length of ORFs to consider. Defaults to 10000.
        overlap (bool, optional):
            Whether to allow overlapping ORFs. Defaults to False.
        include_reverse (bool, optional):
            Whether to include reverse strand ORFs. Defaults to True.
        sort_by_length (bool, optional):
            Whether to sort the results by length in descending order.
            Defaults to True.
        translate (bool, optional):
            Whether to translate the nucleotide sequences to amino acid
            sequences. Defaults to False.
        start_codons (set[str], optional):
            Set of start codons to consider. Defaults to {"ATG"}.
        seq_id (str | None, optional):
            Identifier for the sequence. Defaults to None.

    Returns:
        IntervalResult:
            An object containing found ORF intervals and metadata.

    Raises:
        TypeError:
            If input types are incorrect.
        ValueError:
            If input values are invalid or empty where not allowed.
    """

    # Type and value validation

    # Input: seq
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "find_orfs() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    # Input: min_length & max_length
    if not isinstance(min_length, int):
        raise TypeError(
            "find_orfs() argument 'min_length' must be a non-negative integer"
        )
    if min_length < 0:
        raise ValueError(
            "find_orfs() argument 'min_length' must be non-negative, got "
            + str(min_length)
        )
    if not isinstance(max_length, int):
        raise TypeError(
            "find_orfs() argument 'max_length' must be a non-negative integer"
        )
    if max_length < 0:
        raise ValueError("max_length must be non-negative")
    if max_length < min_length:
        raise ValueError("max_length must be >= min_length")
    # Input: start_codons
    if not start_codons:
        raise ValueError("start_codons cannot be empty")
    validated_start_codons = set()
    for codon in start_codons:
        if not isinstance(codon, str):
            raise TypeError(
                "find_orfs() argument 'start_codons' must contains str , got "
                + type(codon).__name__
            )
        if len(codon) != 3:
            raise ValueError(
                f"Start codon '{codon}' must be exactly 3 nucleotides long"
            )
        codon_upper = codon.upper().replace('U', 'T')
        validated_start_codons.add(codon_upper)
    # Input: seq_id
    if seq_id is not None and not isinstance(seq_id, str):
        raise TypeError(
            "find_orfs() argument 'seq_id' must be str, not "
            + type(seq_id).__name__
        )

    if isinstance(seq, str):
        seq = Sequence(seq, strict=False)
    if seq.is_rna is True:
        seq = seq.reverse_transcribe()

    seq_length = len(seq)
    results = []

    for frame in (0, 1, 2):
        results.extend(
            find_orfs_in_frame(
                seq.sequence, min_length=min_length, max_length=max_length,
                overlap=overlap, strand='+', frame=frame, translate=translate,
                start_codons=validated_start_codons, seq_id=seq_id
            )
        )

    if include_reverse:
        rev_seq = seq.reverse_complement()
        for frame in (0, 1, 2):
            rev_orfs = find_orfs_in_frame(
                rev_seq.sequence, min_length=min_length, max_length=max_length,
                overlap=overlap, strand='-', frame=frame, translate=translate,
                start_codons=validated_start_codons, seq_id=seq_id
                )
            for orf in rev_orfs:
                results.append(
                    SeqInterval(
                        start=seq_length - orf.end,
                        end=seq_length - orf.start,
                        nt_seq=orf.nt_seq, type='ORF',
                        strand='-', frame=-(frame + 1),
                        aa_seq=orf.aa_seq, seq_id=seq_id
                    )
                )

    if sort_by_length:
        results.sort(key=lambda orf: orf.length, reverse=True)

    return IntervalResult(
        intervals=results,
        seq_id=seq_id,
        plot_func=plot_orfs, type="ORF",
        metadata={
            "seq_length": seq_length,
            "sequence": str(seq)
        }
    )


def main():
    from omibio.io.read_fasta import read_fasta
    seqs = read_fasta(r"./examples/data/example_single_short_seq.fasta")
    sequence = seqs["example"]
    res = find_orfs(sequence, translate=True, seq_id='example', min_length=0)
    res.plot(show=True)


if __name__ == "__main__":
    main()



import random
from omibio.sequence import Sequence, Polypeptide
from omibio.io.write_fasta import write_fasta


def random_seq(
    length: int,
    alphabet: str = "ATCG",
    weights: list[float] | None = None,
    seed: int | None = None,
    as_str: bool = False,
    as_polypeptide: bool = False,
    seq_strict: bool = False
) -> Sequence | str | Polypeptide:
    """Generate a random sequence.

    Args:
        length (int):
            Length of the sequence to generate.
        alphabet (str, optional):
            Alphabet to sample from. Defaults to "ATCG".
        weights (list[float] | None, optional):
            Weights for each character in the alphabet. Defaults to None.
        seed (int | None, optional):
            Random seed for reproducibility. Defaults to None.
        as_str (bool, optional):
            If True, return the sequence as a string. Defaults to False.
        seq_strict (bool, optional):
            If True and as_str is False, create a Sequence object in strict
            mode. Defaults to False.

    Raises:
        ValueError: length must be a non-negative integer
        ValueError: Alphabet cannot be empty
        ValueError: Length of 'weights' must match length of 'alphabet'

    Returns:
        Sequence | str: The generated random sequence.
    """

    if length < 0:
        raise ValueError(
            f"Length must be a non-negative integer, got {length}"
        )

    if not alphabet:
        raise ValueError("Alphabet cannot be empty")

    if weights is not None:
        if len(weights) != len(alphabet):
            raise ValueError(
                "Length of 'weights' must match length of 'alphabet'"
            )

    rng = random.Random(seed)

    if weights is not None:
        seq = "".join(rng.choices(alphabet, weights, k=length))
    else:
        seq = "".join(rng.choices(alphabet, k=length))

    if as_polypeptide:
        return Polypeptide(seq, strict=seq_strict)

    return seq if as_str else Sequence(seq, strict=seq_strict)


def random_fasta(
    seq_num: int,
    length: int,
    file_path: str | None = None,
    alphabet: str = "ATCG",
    prefix: str = "Sequence",
    weights: list[float] | None = None,
    seed: int | None = None,
) -> list[str]:
    """Generate a FASTA file containing a random sequence.

    Args:
        file_path (str):
            Path to the output FASTA file.
        seq_num (int):
            Number of sequences to generate.
        length (int):
            Length of each sequence.
        alphabet (str, optional):
            Alphabet to sample from. Defaults to "ATCG".
        prefix (str, optional):
            Prefix for sequence IDs. Defaults to "Sequence".
        weights (list[float] | None, optional):
            Weights for each character in the alphabet. Defaults to None.
        seed (int | None, optional):
            Random seed for reproducibility. Defaults to None.
    """

    rng = random.Random(seed)
    seq_dict = {}

    for i in range(1, seq_num+1):
        seq_seed = rng.randint(0, 2**32 - 1)
        seq_dict[f"{prefix}_{i}"] = random_seq(
            length=length, alphabet=alphabet,
            weights=weights, as_str=True, seed=seq_seed
        )
    if file_path is not None:
        res = write_fasta(file_name=file_path, seqs=seq_dict)
    else:
        res = write_fasta(seqs=seq_dict)
    return res


def main():
    output_path = r"./examples/data/random.fasta"
    random_fasta(output_path, 30, 2000)
    print(output_path)


if __name__ == "__main__":
    main()



from omibio.sequence.sequence import Sequence, Polypeptide
from typing import Literal, Iterable, Mapping
import re
from dataclasses import dataclass


@dataclass
class CleanReportItem:
    orig_name: str
    clean_name: str | None = None
    orig_len: int = 0
    clean_len: int = 0
    removed: bool = False
    gap_policy: str | None = None
    illegal_removed: int = 0
    illegal_replaced: int = 0
    name_changed: bool = False
    reason: str = ""


class CleanReport:

    def __init__(self) -> None:
        self._records: list[CleanReportItem] = []

    def add(self, item: CleanReportItem):
        if not isinstance(item, CleanReportItem):
            raise TypeError(
                "add() argument 'item' must be CleanReportItem, got "
                + type(item).__name__
            )
        self._records.append(item)

    @property
    def records(self) -> list[CleanReportItem]:
        return self._records

    @property
    def kept(self):
        return [r for r in self._records if not r.removed]

    @property
    def removed(self):
        return [r for r in self._records if r.removed]

    def summary(self):
        return {
            "total": len(self._records),
            "kept": len(self.kept),
            "removed": len(self.removed)
        }


VALID_BASES = {
    "A", "T", "C", "G", "U",
    "N", "R", "Y", "K", "M", "B", "V", "D", "H", "S", "W"
}
WHITESPACE_RE = re.compile(r"\s+")
INVALID_NAME_CHAR_RE = re.compile(r"[^ A-Za-z0-9_.-]")
ALIG_SYMBOL_RE = re.compile(r"-+")


def clean(
    seqs: Mapping[str, str | Sequence | Polypeptide],
    name_policy: Literal["keep", "id_only", "underscores"] = "keep",
    gap_policy: Literal["keep", "remove", "collapse"] = "keep",
    strict: bool = False,
    min_len: int = 10,
    max_len: int = 100_000,
    normalize_case: bool = True,
    remove_illegal: bool = False,
    allowed_bases: Iterable[str] | None = None,
    remove_empty: bool = True,
    as_str: bool = False,
    as_polypeptide: bool = False,
    report: bool = False
) -> (
    dict[str, str | Sequence | Polypeptide]
    | tuple[dict[str, str | Sequence | Polypeptide], CleanReport]
):
    """Clean sequences according to specified policies.

    Args:
        seqs (dict[str, str  |  Sequence]):
            Dictionary of sequence names to sequences
            (as strings or Sequence objects).
        name_policy (Literal["keep", "id_onbly", "underscores"], optional):
            Policy for cleaning sequence names. Defaults to "keep".
        gap_policy (Literal["keep", "remove", "collapse"], optional):
            Policy for handling gaps in sequences. Defaults to "keep".
        strict (bool, optional):
            If True, raises an error on illegal characters. Defaults to False.
        min_len (int, optional):
            Minimum length for sequences to keep. Defaults to 10.
        max_len (int, optional):
            Maximum length for sequences to keep. Defaults to 100,000.
        normalize_case (bool, optional):
            If True, converts sequences to uppercase. Defaults to True.
        remove_illegal (bool, optional):
            If True, removes illegal characters instead of replacing
            them with 'N'. Defaults to False.
        allowed_bases (set[str] | None, optional):
            Set of allowed bases. Defaults to VALID_BASES.
        remove_empty (bool, optional):
            If True, removes sequences that are empty or contain only 'N's.
            Defaults to True.
        as_str (bool, optional):
            If True, returns cleaned sequences as strings.
            If False, returns them as Sequence objects. Defaults to True.
        report (bool, optional):
            If True, generates a cleaning report. Defaults to False.

    Raises:
        ValueError:
            If input values are invalid.
        TypeError:
            If input types are incorrect.

    Returns:
        dict[str, str | Sequence] |
        tuple[dict[str, str | Sequence], CleanReport]:
            Cleaned sequences, optionally with a cleaning report.
    """

    if allowed_bases is None:
        allowed_bases = VALID_BASES
    else:
        allowed_bases = set(allowed_bases)
    for base in allowed_bases:
        if not isinstance(base, str) or len(base) != 1:
            raise ValueError(
                "clean() argument 'allowed_bases' must be an iterable "
                "of single-character strings."
            )
    if not isinstance(seqs, dict):
        raise TypeError(
            f"clean() argument 'seqs' must be dict, got {type(seqs).__name__}"
        )
    if name_policy not in {"keep", "id_only", "underscores"}:
        raise ValueError(
            "clean() argument 'name_policy' must be one of: "
            "'keep', 'id_only', 'underscores'"
        )
    if gap_policy not in {"keep", "remove", "collapse"}:
        raise ValueError(
            "clean() argument 'gap_policy' must be one of: "
            "'keep', 'remove', 'collapse'"
        )
    if not isinstance(min_len, int):
        raise TypeError(
            "clean() argument 'min_len' must be int, got "
            + type(min_len).__name__
        )
    if not isinstance(max_len, int):
        raise TypeError(
            "clean() argument 'max_len' must be int, got "
            + type(max_len).__name__
        )
    if min_len < 0 or max_len < 0:
        raise ValueError(
            "clean() argument 'min_len' and 'max_len' must be "
            "non-negative numbers"
        )
    if min_len > max_len:
        raise ValueError(
            "clean() argument 'min_len' cannot be larger than 'max_len'"
        )
    cleaned_seqs = {}
    if report:
        clean_report = CleanReport()

    # ---------------- name processing ----------------
    def process_name(name) -> str:
        name = WHITESPACE_RE.sub(" ", name.strip())

        if name_policy == "id_only":
            name = name.split(" ", 1)[0]

        name = INVALID_NAME_CHAR_RE.sub("_", name)

        if name_policy == "underscores":
            name = name.replace(" ", "_")

        if not name:
            name = "unnamed"

        return name

    # ---------------- sequence cleaning ----------------
    def process_seq(seq: str, item: CleanReportItem) -> str:

        pure_seq = WHITESPACE_RE.sub("", seq)
        item.orig_len = len(pure_seq)

        cleaned = pure_seq
        if normalize_case:
            cleaned = cleaned.upper()

        match gap_policy:
            case "remove":
                item.gap_policy = "remove"
                cleaned = cleaned.replace("-", "")
            case "collapse":
                item.gap_policy = "collapse"
                cleaned = ALIG_SYMBOL_RE.sub("-", cleaned)
            case _:
                item.gap_policy = "keep"

        if strict:
            for i, base in enumerate(cleaned):
                if base not in allowed_bases and base != "-":
                    raise ValueError(
                        f"Illegal character {base} in sequence at position {i}"
                    )
        else:
            illegal_removed = 0
            illegal_replaced = 0
            new_seq = []

            for base in cleaned:
                if base in allowed_bases or base == "-":
                    new_seq.append(base)
                else:
                    if remove_illegal:
                        illegal_removed += 1
                    else:
                        new_seq.append("N")
                        illegal_replaced += 1

            cleaned = "".join(new_seq)
            item.illegal_removed = illegal_removed
            item.illegal_replaced = illegal_replaced

        if remove_empty and set(cleaned) <= {"N", "-"}:
            item.reason = "empty_or_N_only"
            return ""

        if not cleaned or not (min_len <= len(cleaned) <= max_len):
            item.reason = "length_filter"
            return ""

        return cleaned

    # ---------------- main loop ----------------
    for raw_name, raw_seq in seqs.items():
        if not isinstance(raw_name, str):
            raise TypeError(
                "Sequence name must be a string, got "
                + type(raw_name).__name__
            )
        if not isinstance(raw_seq, (str, Sequence)):
            raise TypeError(
                f"Sequence '{raw_name}' must be a string or Sequence, "
                f"got {type(raw_seq).__name__}"
            )

        item = CleanReportItem(orig_name=raw_name)

        cleaned_name = process_name(raw_name)

        cleaned = process_seq(str(raw_seq), item)

        if not cleaned:
            item.removed = True
            item.clean_len = 0
            if report:
                clean_report.add(item)
            continue

        if cleaned_name in cleaned_seqs:
            count = 1
            new_name = f"{cleaned_name}_{count}"
            while new_name in cleaned_seqs:
                count += 1
                new_name = f"{cleaned_name}_{count}"
            cleaned_name = new_name

        item.clean_name = cleaned_name
        item.clean_len = len(cleaned)

        if cleaned_name != raw_name:
            item.name_changed = True

        cleaned_seq: str | Sequence | Polypeptide
        if not as_str:
            if as_polypeptide:
                cleaned_seq = Polypeptide(cleaned)
            else:
                cleaned_seq = Sequence(cleaned)
        else:
            cleaned_seq = cleaned

        cleaned_seqs[cleaned_name] = cleaned_seq

        if report:
            clean_report.add(item)

    if not report:
        return cleaned_seqs
    else:
        return cleaned_seqs, clean_report


def write_report(out_path: str, report: CleanReport) -> None:
    """Write a cleaning report to a text file.

    Args:
        out_path (str): Path to the output report file.
        report (CleanReport): The cleaning report to write.
    Raises:
        TypeError:
            If input types are incorrect.
    """

    if not isinstance(out_path, str):
        raise TypeError(
            "write_report() argument 'out_path' must be str, got "
            + type(out_path).__name__
        )
    if not isinstance(report, CleanReport):
        raise TypeError(
            "write_report() argument 'report' must be CleanReport, got "
            + type(report).__name__
        )

    lines = []

    # --- Summary ---
    lines.append("=== Clean Report Summary ===")
    summary = report.summary()
    lines.append(f"Total sequences: {summary['total']}")
    lines.append(f"Kept: {summary['kept']}")
    lines.append(f"Removed: {summary['removed']}")
    lines.append("")

    # --- Kept sequences table ---
    lines.append("=== Cleaned Sequences ===")
    headers = [
        "orig_name", "clean_name",
        "orig_len", "clean_len",
        "name_changed", "gap_policy",
        "illegal_removed", "illegal_replaced"
    ]

    # Prepare rows
    rows = []
    for r in report.kept:
        rows.append([
            r.orig_name,
            r.clean_name,
            str(r.orig_len),
            str(r.clean_len),
            "yes" if r.name_changed else "no",
            r.gap_policy if r.gap_policy else "",
            str(r.illegal_removed),
            str(r.illegal_replaced)
        ])

    # Column widths
    col_widths = [
        max(
            len(headers[i]), max((len(row[i]) for row in rows), default=0)
        ) for i in range(len(headers))
    ]

    # Header row
    header_line = "  ".join(
        headers[i].ljust(col_widths[i]) for i in range(len(headers))
    )
    lines.append(header_line)
    lines.append("-" * len(header_line))

    # Data rows
    for row in rows:
        lines.append(
            "  ".join(row[i].ljust(col_widths[i]) for i in range(len(headers)))
        )
    lines.append("")

    # --- Removed sequences ---
    lines.append("=== Removed Sequences ===")
    lines.append("orig_name          reason")
    lines.append("-" * 50)
    for r in report.removed:
        reason = getattr(r, "reason", "removed")
        lines.append(f"{r.orig_name.ljust(18)} {reason}")

    # Write to file
    with open(out_path, "w", encoding="utf-8") as f:
        f.write("\n".join(lines))


def main():
    from omibio.io import write_fasta, read_fasta
    input_path = "./examples/data/example_dirty.fasta"
    output_path = "./examples/output/clean_fasta_output.fasta"
    report_path = "./examples/output/clean_report.txt"

    seqs = read_fasta(input_path, strict=False).seq_dict()

    cleaned_seqs, report = clean(
        seqs, name_policy="id_only", gap_policy="collapse",
        report=True, remove_illegal=True
    )

    write_fasta(file_name=output_path, seqs=cleaned_seqs)
    print(f"Cleaned: {output_path}")

    write_report(report_path, report)
    print(f"Report: {report_path}")


if __name__ == "__main__":
    main()



from omibio.sequence import Sequence
from collections import defaultdict

IUPAC_CODES = {
    frozenset({'A', 'G'}): 'R', frozenset({'C', 'T'}): 'Y',
    frozenset({'G', 'C'}): 'S', frozenset({'A', 'T'}): 'W',
    frozenset({'G', 'T'}): 'K', frozenset({'A', 'C'}): 'M',
    frozenset({"C", "G", "T"}): "B", frozenset({"A", "G", "T"}): "D",
    frozenset({"A", "C", "T"}): "H", frozenset({"A", "C", "G"}): "V"
}


def find_consensus(
    seq_list: list[str | Sequence],
    as_str: bool = False,
    gap_chars: str = "-?.",
    as_rna: bool = False
) -> str | Sequence:
    """Finds the consensus sequence from a list of sequences.

    Args:
        seq_list (list[str  |  Sequence]):
            list of sequences to find consensus from.
        as_str (bool, optional):
            whether to return the consensus as a string.
            Defaults to False.
        gap_chars (str, optional):
            characters to consider as gaps. Defaults to "-?.".
        as_rna (bool, optional):
            whether to return the consensus as RNA sequence.
            Defaults to False.

    Raises:
        TypeError: If seq_list is not a list.
        ValueError: If sequences are not of the same length.

    Returns:
        str | Sequence: The consensus sequence as a string or Sequence object.
    """

    if not seq_list:
        return ""
    if not isinstance(seq_list, list):
        raise TypeError(
            "find_consensus() argument 'seq_list' must be list, got "
            + type(seq_list).__name__
        )
    seq_list = [str(s).upper() for s in seq_list]

    lengths = set(map(len, seq_list))
    if len(lengths) != 1:
        raise ValueError("All sequences must be of the same length")

    consensus_list = []

    for i in range(lengths.pop()):
        base_scores: defaultdict[str, int] = defaultdict(int)
        for seq in seq_list:
            base = seq[i].replace("U", "T")
            if base in gap_chars:
                continue
            base_scores[base] += 1

        if not base_scores:
            consensus_list.append("N")
            continue

        max_score = max(base_scores.values())
        top_base = [
            b for b, score in base_scores.items() if score == max_score
        ]
        if len(top_base) == 1:
            consensus_list.append(top_base[0])
        else:
            consensus_list.append(IUPAC_CODES.get(frozenset(top_base), "N"))

    consensus_seq = "".join(consensus_list)
    consensus = (
        Sequence(consensus_seq, strict=False) if not as_rna
        else Sequence(consensus_seq.replace("T", "U"), strict=False)
    )

    return consensus_seq if as_str else consensus


def main():
    from omibio.io.read_fasta import read_fasta
    input_file = r"./examples/data/example_short_seqs.fasta"
    sequences = read_fasta(input_file).seqs()
    consensus = find_consensus(sequences, as_rna=False)
    print(consensus)


if __name__ == "__main__":
    main()



from omibio.bio import SeqInterval, IntervalResult
from omibio.sequence.sequence import Sequence
from omibio.viz.plot_sliding_gc import plot_sliding_gc


def sliding_gc(
    seq: Sequence | str,
    window: int = 100,
    step: int = 10,
    seq_id: str | None = None
) -> IntervalResult:
    """Calculate GC content in a sliding window manner.

    Args:
        seq (Sequence | str):
            input sequence
        window (int, optional):
            window size. Defaults to 100.
        step (int, optional):
            step size. Defaults to 10.
        seq_id (str | None, optional):
            an optional identifier for the sequence. Defaults to None.

    Raises:
        ValueError: if window or step is not positive.
        TypeError: if seq is not Sequence or str.

    Returns:
        AnalysisResult:
            GC content analysis result object.
    """

    if not seq:
        return IntervalResult(
            intervals=[], seq_id=seq_id, type="sliding_gc",
            metadata={
                "seq_length": 0,
                "sequence": ""
            }
        )
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "sliding_gc() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    if window <= 0 or step <= 0:
        raise ValueError("window and step should be positive numbers")
    if isinstance(seq, Sequence):
        seq = str(seq)

    n = len(seq)
    seq = seq.upper()

    if window >= n:
        gc_count = sum(1 for b in seq if b in 'GC')
        gc_percent = round((gc_count / n), 3)
        return IntervalResult(
            intervals=[
                SeqInterval(
                    start=0, end=n, gc=gc_percent, type="GC", seq_id=seq_id
                )
            ],
            seq_id=seq_id, type="sliding_gc", plot_func=plot_sliding_gc,
            metadata={
                "seq_length": n,
                "sequence": str(seq)
            }
        )

    is_gc = [1 if b in 'GC' else 0 for b in seq]

    gc_count = sum(is_gc[:window])
    gc_list = [
        SeqInterval(
            start=0, end=window,
            gc=round((gc_count / window), 3), type="GC", seq_id=seq_id
        )
    ]

    for i in range(step, n - window + 1, step):
        gc_count -= sum(is_gc[i-step: i])
        gc_count += sum(is_gc[i+window-step: i+window])

        gc_percent = round((gc_count / window), 3)
        gc_list.append(
            SeqInterval(
                start=i, end=i+window,
                gc=gc_percent, type="GC", seq_id=seq_id
                )
        )

    return IntervalResult(
        intervals=gc_list,
        seq_id=seq_id,
        plot_func=plot_sliding_gc,
        type="sliding_gc",
        metadata={
            "seq_length": n,
            "sequence": str(seq)
        }
    )


def main():
    from omibio.io.read_fasta import read_fasta
    seq = read_fasta(
        "./examples/data/example_single_long_seq.fasta"
    )["example"]
    gc_res = sliding_gc(seq)
    print(gc_res)
    gc_res.plot(show=True)


if __name__ == "__main__":
    main()



import click
from omibio.io import read_fasta
from omibio.cli.gc_cli import gc_group
from omibio.analysis import sliding_gc
from omibio.sequence import Polypeptide
import csv


@gc_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "--window",
    type=int,
    default=100,
    help="window size. Defaults to 100."
)
@click.option(
    "--step",
    type=int,
    default=10,
    help="step size. Defaults to 10."
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None,
    help="Write details to a file in csv format"
)
@click.option(
    "--summary", "-s",
    is_flag=True
)
def window_gc(
    source: str,
    window: int,
    step: int,
    output: str | None,
    summary: bool
):
    """
    Calculate and plot sliding window GC content for sequences in a FASTA file.
    """

    entries = read_fasta(source)

    if summary:
        for entry in entries:
            gc_vals = []
            seq = entry.seq
            if isinstance(seq, Polypeptide):
                raise TypeError(
                    "GC content can only be calculated for nucleotide "
                    "sequences, got Polypeptide"
                )
            result = sliding_gc(
                seq, window=window, step=step, seq_id=entry.seq_id
            )

            for interval in result:
                if (gc_val := interval.gc) is not None:
                    gc_vals.append(gc_val)

            average = sum(gc_vals) / len(gc_vals)
            click.echo(
                f"{entry.seq_id}\tmean={average:.2f}%\t"
                f"max={max(gc_vals)}%\tmin={min(gc_vals)}%"
            )
    else:
        rows = [["seq_id", "start", "end", "gc"]]
        for entry in entries:
            seq = entry.seq
            if isinstance(seq, Polypeptide):
                raise TypeError(
                    "GC content can only be calculated for nucleotide "
                    "sequences, got Polypeptide"
                )
            for itv in list(sliding_gc(
                seq, window=window, step=step, seq_id=entry.seq_id
            )):
                rows.append([
                    str(itv.seq_id), str(itv.start), str(itv.end), str(itv.gc)
                ])
        if output is not None:
            with open(output, "w", newline="", encoding="utf-8") as f:
                writer = csv.writer(f, delimiter="\t")
                writer.writerows(rows)
                click.echo(f"Written to {output}")
        else:
            for row in rows:
                click.echo("\t".join(row))



from dataclasses import dataclass, field
from omibio.bio.analysis_result import AnalysisResult
from typing import Iterator, Iterable


@dataclass
class KmerResult(AnalysisResult):
    """lass to hold kmer counting results. is a subclass of AnalysisResult.

    Args:
        AnalysisResult:
            Base class for analysis results.

    Raises:
        TypeError:
            If the input types are incorrect.
    """

    k: int = field(default_factory=int)
    counts: dict[str, int] = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.counts, dict):
            raise TypeError(
                "KmerResult argument 'counts' must be dict, got "
                + type(self.counts).__name__
            )

    def items(self) -> Iterable[tuple[str, int]]:
        """Return an iterator over the (kmer, count) pairs."""
        return self.counts.items()

    def keys(self) -> Iterable[str]:
        """Return an iterator over the kmers."""
        return self.counts.keys()

    def values(self) -> Iterable[int]:
        """Return an iterator over the counts."""
        return self.counts.values()

    def __len__(self) -> int:
        return len(self.counts)

    def __iter__(self) -> Iterator[str]:
        return iter(self.counts)

    def __getitem__(self, key: str) -> int:
        return self.counts[key]

    def __repr__(self) -> str:
        return (
            f"KmerResult(counts={self.counts!r}, "
            f"seq_id={self.seq_id!r}, type={self.type!r})"
        )

    def __str__(self) -> str:
        return str(self.counts)


def main():
    ...


if __name__ == "__main__":
    main()



from omibio.sequence.polypeptide import Polypeptide
from omibio.sequence.sequence import Sequence
from omibio.sequence.seq_utils import (
    clean, CleanReport, CleanReportItem,
    complement, reverse_complement,
    random_seq, random_fasta,
    shuffle_seq,
    transcribe, reverse_transcribe,
    translate_nt
)


__all__ = [
    "Polypeptide",
    "Sequence",
    "clean",
    "CleanReport",
    "CleanReportItem",
    "complement",
    "reverse_complement",
    "random_seq",
    "random_fasta",
    "shuffle_seq",
    "transcribe",
    "reverse_transcribe",
    "translate_nt"
]



import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.axes import Axes
from omibio.bio import SeqInterval
from omibio.bio import IntervalResult


def plot_orfs(
    orfs: list[SeqInterval] | IntervalResult,
    seq_length: int | None = None,
    ax: Axes | None = None,
    show: bool = False
) -> Axes:
    """Plot Open Reading Frames (ORFs) along a sequence.

    Args:
        orfs (list[SeqInterval] | IntervalResult):
            List of ORF intervals or an IntervalResult object.
        seq_length (int | None, optional):
            Length of the sequence. Required if 'orfs' is a list.
            Defaults to None.
        ax (Axes | None, optional):
            Matplotlib Axes to plot on. If None, a new figure and axes
            will be created. Defaults to None.
        show (bool, optional):
            Whether to display the plot immediately. Defaults to False.

    Raises:
        TypeError:
            If 'orfs' is a list and 'seq_length' is not provided.

    Returns:
        Axes:
            The Matplotlib Axes object containing the plot.
    """

    if ax is None:
        _, ax = plt.subplots(figsize=(9, 3))

    if seq_length is None:
        if isinstance(orfs, IntervalResult):
            seq_length = orfs.metadata["seq_length"]
        else:
            raise TypeError(
                "plot_orfs() argument: 'seq_length' must be provided if"
                "'orfs' is a list"
            )

    frame_y = {
        '+1': 5, '+2': 4, '+3': 3,
        '-1': 2, '-2': 1, '-3': 0
    }
    bar_height = 0.4
    positive_color = "#4D84DC"
    negative_color = "#E14040"

    for i in frame_y.values():
        ax.axhline(y=i, color="#C8C7CE", linestyle="--")

    for orf in orfs:
        frame_key = f"+{orf.frame}" if orf.frame > 0 else str(orf.frame)
        y = frame_y[frame_key]
        color = positive_color if orf.strand == '+' else negative_color

        ax.broken_barh(
            [(orf.start, orf.length)],
            (y - bar_height/2, bar_height),
            facecolors=color, zorder=3
        )

    ax.set_ylim(-1, 7)
    ax.set_xlim(0, seq_length)
    ax.set_yticks(list(frame_y.values()))
    ax.set_yticklabels(list(frame_y.keys()))
    ax.set_xlabel("Sequence Position")
    ax.set_ylabel("Reading Frame")
    ax.set_title("ORF Distribution")

    legend_elements = [
        Patch(facecolor=positive_color, label="Positive strand (+)"),
        Patch(facecolor=negative_color, label="Negative strand (-)")
    ]
    ax.legend(handles=legend_elements, loc="upper right")
    ax.grid(True, linestyle='--', alpha=0.5)

    if show:
        plt.show()
    return ax


def main():
    from omibio.analysis.find_orfs import find_orfs
    from omibio.io.read_fasta import read_fasta
    seq = read_fasta(
        "./examples/data/example_single_long_seq.fasta"
    )["example"]
    orfs = find_orfs(seq, min_length=100)
    plot_orfs(orfs)
    plt.show()


if __name__ == "__main__":
    main()



import click
from omibio.io import read_fasta
from omibio.cli.gc_cli import gc_group
from omibio.sequence import Polypeptide
from typing import TextIO
import csv


@gc_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "-o", "--output",
    type=click.Path(),
    help="Output file path."
)
def compute(source: TextIO, output: str | None) -> None:
    """Calculate the GC content of a sequence from a FASTA file."""

    seqs = read_fasta(source, strict=False).seq_dict()
    lines = []

    for name, seq in seqs.items():
        if isinstance(seq, Polypeptide):
            raise TypeError(
                f"Cannot calculate gc for amino acid sequences: {name}"
            )
        gc_val = seq.gc_content(percent=True)
        lines.append((name, gc_val))

    if output is None:
        for line in lines:
            click.echo("\t".join(map(str, line)))
    else:
        with open(output, "w", newline="", encoding="utf-8") as f:
            writer = csv.writer(f)
            writer.writerow(["seq_id", "gc"])
            writer.writerows(lines)
        click.echo(f"Result written to {output}")



from pathlib import Path
from typing import TYPE_CHECKING, Iterator, TextIO, cast
from os import PathLike
import warnings
if TYPE_CHECKING:
    from omibio.bio import SeqCollections, SeqEntry

VALID_NT = {
    "A", "T", "U", "C", "G", "R", "Y", "K",
    "M", "B", "V", "D", "H", "S", "W", "N"
}
VALID_AA = {
    "A", "C", "D", "E", "F", "G", "H", "I", "K", "L", "M",
    "N", "P", "Q", "R", "S", "T", "V", "W", "Y", "X", "*"
}


class FastaFormatError(Exception):

    def __init__(self, message):
        super().__init__(message)


def read_fasta_iter(
    source: str | TextIO | PathLike,
    strict: bool = False,
    warn: bool = True,
    output_strict: bool = False,
    skip_invalid_seq: bool = False,
) -> Iterator["SeqEntry"]:
    """Parse a FASTA file and return a Generator.

    Raises:
        FileNotFoundError:
            If the specified file is not found.
        FastaFormatError:
            If the FASTA format is invalid.
        FastaFormatError:
            If the sequence contains invalid characters.
        FastaFormatError:
            If the sequence name is missing.
        FastaFormatError:
            If the sequence is missing.

    Yields:
        SeqEntry:
            SeqEntry objects for each sequence in the FASTA file.
    """

    from omibio.sequence import Sequence, Polypeptide
    from omibio.bio import SeqEntry

    if hasattr(source, "read"):
        fh = cast(TextIO, source)
        file_name = "<stdin>"
        faa = False
    else:
        file_path = Path(source)
        if not file_path.exists():
            raise FileNotFoundError(f"File '{source}' not found.")

        suffix = file_path.suffix.lower()
        if suffix not in {".faa", ".fa", ".fasta", ".fna"}:
            raise FastaFormatError(
                f"Invalid format to read: {suffix}"
            )
        faa = (suffix == ".faa")
        file_name = str(file_path)
        fh = open(file_path, "r")

    current_name = None
    current_seq: list[str] = []
    allowed_set = VALID_AA if faa else VALID_NT

    def push_entry():
        if current_name is None:
            return
        if not current_seq:
            msg = f"Sequence missing for {current_name}"
            if strict:
                raise FastaFormatError(msg)
            elif warn:
                warnings.warn(msg + ", skip record")
            return

        seq_str = "".join(current_seq)
        current_seq.clear()
        if faa:
            seq_obj = Polypeptide(seq_str, strict=output_strict)
        else:
            seq_obj = Sequence(seq_str, strict=output_strict)

        return SeqEntry(seq=seq_obj, seq_id=current_name, source=file_name)

    try:
        for i, line in enumerate(fh, start=1):
            line = line.split("#", 1)[0].strip()
            if not line:
                continue

            if line.startswith(">"):
                entry = push_entry()
                if entry:
                    yield entry

                current_name = line[1:].strip()

                if not current_name:
                    if strict:
                        raise FastaFormatError(
                            f"Sequence name missing in line {i}"
                        )
                    elif warn:
                        warnings.warn(
                            f"Sequence name missing in line {i}, "
                            "skip record"
                        )
                    current_name = None
                    current_seq.clear()
                    continue

            else:
                if current_name is None:
                    continue
                # Store sequence.
                line = line.upper()
                skip_record = False
                if strict or warn or skip_invalid_seq:
                    if any(c not in allowed_set for c in line):
                        if strict:
                            raise FastaFormatError(
                                f"Invalid sequence in line {i}: {line}"
                            )
                        elif warn:
                            warnings.warn(
                                f"Invalid sequence in line {i}: {line}, "
                                f"{'skip' if skip_invalid_seq else 'invalid'} "
                                "record"
                            )
                        if skip_invalid_seq:
                            skip_record = True
                if skip_record:
                    current_name = None
                    current_seq.clear()
                    continue

                current_seq.append(line)

        # Store the last sequence when the file ends
        entry = push_entry()
        if entry:
            yield entry
    finally:
        if not hasattr(source, "read"):
            fh.close()


def read_fasta(
    source: str | TextIO | PathLike,
    strict: bool = False,
    output_strict: bool = False,
    warn: bool = True,
    skip_invalid_seq: bool = False
) -> "SeqCollections":
    """Parse a FASTA file and return a SeqCollections object.

    Args:
        source (str | TextIO | PathLike):
            Path to the FASTA file or a file-like object.
        strict (bool, optional):
            Whether to raise errors on invalid sequences. Defaults to False.
        output_strict (bool, optional):
            Whether to enforce strictness on the output sequences.
            Defaults to False.
        warn (bool, optional):
            Whether to issue warnings on invalid sequences. Defaults to True.
        skip_invalid_seq (bool, optional):
            Whether to skip invalid sequences. Defaults to False.

    Returns:
        SeqCollections:
            A SeqCollections object containing the parsed sequences.
    """

    from omibio.bio import SeqCollections

    entries = []
    for entry in read_fasta_iter(
        source,
        strict=strict,
        output_strict=output_strict,
        warn=warn,
        skip_invalid_seq=skip_invalid_seq
    ):
        entries.append(entry)

    if hasattr(source, "read"):
        file_name = "<stdin>"
    else:
        file_name = str(source)

    return SeqCollections(entries=entries, source=file_name)


def main():
    input_path = r"./examples/data/example_lots_of_seqs.fasta"
    for entry in read_fasta(
        input_path, strict=False, skip_invalid_seq=True, warn=True
    ):
        print(entry.source)


if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from omibio.bio import SeqInterval, IntervalResult


def plot_motifs(
    motifs: list[SeqInterval] | IntervalResult,
    seq_length: int | None = None,
    ax: Axes | None = None,
    show: bool = False
) -> Axes:
    """Plot motif occurrences along a sequence.

    Args:
        motifs (list[SeqInterval] | IntervalResult):
            List of motif intervals or an IntervalResult object.
        seq_length (int | None, optional):
            Length of the sequence. Required if 'motifs' is a list.
            Defaults to None.
        ax (Axes | None, optional):
            Matplotlib Axes to plot on. If None, a new figure and axes
            will be created. Defaults to None.
        show (bool, optional):
            Whether to display the plot immediately. Defaults to False.

    Raises:
        TypeError:
            If 'motifs' is a list and 'seq_length' is not provided.

    Returns:
        Axes:
            The Matplotlib Axes object containing the plot.
    """

    if ax is None:
        ax = plt.subplots(figsize=(9, 3))[1]
    if seq_length is None:
        if isinstance(motifs, IntervalResult):
            seq_length = motifs.metadata["seq_length"]
        else:
            raise TypeError(
                "plot_motifs() argument: 'seq_length' must be provided if"
                "'motifs' is a list"
            )

    strand_y = {"+": 2, "-": 1}
    color_mp = {"+": "#4D84DC", "-": "#E14040"}
    ax.axhline(y=2, color="#4D84DC", linestyle="-", zorder=0)
    ax.axhline(y=1, color="#E14040", linestyle="-", zorder=0)
    handle_plus = ax.scatter(
        [], [], color=color_mp["+"], marker=">", label="Positive strand (+)"
    )
    handle_minus = ax.scatter(
        [], [], color=color_mp["-"], marker="<", label="Negative strand (-)"
    )

    for motif in motifs:
        color = color_mp[motif.strand]
        marker = ">" if motif.strand == "+" else "<"

        ax.scatter(
            x=motif.start, y=strand_y[motif.strand], s=70,
            facecolors=color, marker=marker, zorder=3
        )

    ax.set_ylim(0, 3)
    ax.set_xlim(-1, seq_length+1)
    ax.set_yticks([2, 1])
    ax.set_yticklabels(["+", "-"])
    ax.set_xlabel("Sequence Position")
    ax.set_ylabel("Strand")
    ax.set_title("Motif Distribution")
    ax.grid(True, linestyle='--', alpha=0.5)
    ax.legend(handles=[handle_plus, handle_minus])

    if show:
        plt.show()
    return ax


def main():
    from omibio.analysis import find_motifs
    from omibio.io import read_fasta

    seq = read_fasta(
        r"./examples/data/example_single_short_seq.fasta"
    )["example"]
    res = find_motifs(seq, "ACT", include_reverse=True)
    plot_motifs(res, show=True)


if __name__ == "__main__":
    main()



from omibio.sequence.sequence import Sequence
from omibio.bio import SeqInterval, IntervalResult


def find_palindrome(
    seq: Sequence | str,
    min_len: int = 4,
    max_len: int = 12,
    seq_id: str | None = None
) -> IntervalResult:
    """Find palindromic sequences in a given sequence.

    Args:
        seq (Sequence | str):
            Input sequence to search for palindromes.
        min_len (int, optional):
            Minimum length of palindromic sequences to find. Defaults to 4.
        max_len (int, optional):
            Maximum length of palindromic sequences to find. Defaults to 12.
        seq_id (str | None, optional):
            Identifier for the sequence. Defaults to None.

    Raises:
        TypeError:
            If the input sequence is not of type Sequence or string.
        TypeError:
            If min_len or max_len is not of type int.
        ValueError:
            If min_len is larger than the length of the sequence.
        ValueError:
            If min_len is larger than max_len.

    Returns:
        IntervalResult:
            An object containing found palindrome intervals and metadata.
    """

    if not seq:
        return IntervalResult(
            intervals=[], seq_id=seq_id, type="palindrome",
            metadata={
                "seq_length": 0,
                "sequence": ""
            }
        )
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "find_palindrome() argument 'seq' must be Sequence or str, got "
            + type(seq).__name__
        )
    if not isinstance(min_len, int) or not isinstance(max_len, int):
        raise TypeError(
            "find_palindrome() argument 'min_len' and 'max_len must be int."
        )

    n = len(seq)

    if min_len > n:
        raise ValueError(
            f"find_palindrome() argument 'min_len' {min_len}"
            f"cannot be larger than the length of sequence {n}"
        )
    if max_len < min_len:
        raise ValueError(
            f"find_palindrome() argument 'min_len' {min_len} "
            f"cannot be larger than 'max_len' {max_len}"
        )

    if isinstance(seq, str):
        seq = Sequence(seq)

    rev_seq = seq.reverse_complement()
    results = []

    start = (max_len) if max_len % 2 == 0 else (max_len - 1)

    for length in range(start, min_len - 1, -2):

        for i in range(0, n - length + 1):
            candidate = seq[i: i+length]
            if candidate == rev_seq[n-i-length: n-i]:
                results.append(
                    SeqInterval(
                        start=i, end=i+length, nt_seq=str(candidate),
                        type='palindrome'
                    )
                )

    return IntervalResult(
        intervals=results, seq_id=seq_id, type="palindrome",
        metadata={
            "seq_length": n,
            "sequence": str(seq)
        }
    )


def main():
    seq = Sequence("GAAUUTGG")
    res = find_palindrome(seq)
    print(list(res))


if __name__ == "__main__":
    main()



import click
from omibio.cli import cli
from omibio.sequence.seq_utils.random_seq import random_fasta as rf


@cli.command()
@click.argument("output", type=str, required=False)
@click.option(
    "-n", "--number",
    type=int,
    default=1,
    help="Number of random sequences to generate (default: 1)."
)
@click.option(
    "-p", "--prefix",
    type=str,
    default="random_seq",
    help="Prefix for sequence IDs (default: 'random_seq')."
)
@click.option(
    "-l", "--length",
    type=int,
    default=100,
    help="Length of sequences"
)
@click.option(
    "--alphabet", "-a",
    type=str,
    default="ATGC",
    help="Alphabet to sample from (default: ATGC)."
)
@click.option(
    "--seed",
    type=int,
    help="Random seed for reproducibility."
)
def random_fasta(
    output: str,
    length: int,
    number: int,
    prefix: str,
    alphabet: str,
    seed: int | None
) -> None:
    """Generate random nucleotide sequence(s) and output in FASTA format."""

    res = rf(
        file_path=output, seq_num=number, length=length,  alphabet=alphabet,
        prefix=prefix, seed=seed
    )
    if output is not None:
        click.echo(f"Success: file writed to {output}")
    else:
        for line in res:
            click.echo(line)



from collections import Counter, defaultdict


class Polypeptide:
    """
    Class representing a polypeptide (amino acid) sequence.
    """

    VALID_AA = {
        "A", "R", "N", "D", "C", "Q", "E", "G", "H", "I",
        "L", "K", "M", "F", "P", "S", "T", "W", "Y", "V"
    }

    AA_MASS = {
        "A": 71.03711, "C": 103.00919, "D": 115.02694, "E": 129.04259,
        "F": 147.06841, "G": 57.02146, "H": 137.05891, "I": 113.08406,
        "K": 128.09496, "L": 113.08406, "M": 131.04049, "N": 114.04293,
        "P": 97.05276, "Q": 128.05858, "R": 156.10111, "S": 87.03203,
        "T": 101.04768, "V": 99.06841, "W": 186.07931, "Y": 163.06333,
    }

    RESIDUE_FORMULA = {
        "A": {"C": 3, "H": 5, "N": 1, "O": 1},
        "R": {"C": 6, "H": 12, "N": 4, "O": 1},
        "N": {"C": 4, "H": 6, "N": 2, "O": 2},
        "D": {"C": 4, "H": 5, "N": 1, "O": 3},
        "C": {"C": 3, "H": 5, "N": 1, "O": 1, "S": 1},
        "E": {"C": 5, "H": 7, "N": 1, "O": 3},
        "Q": {"C": 5, "H": 8, "N": 2, "O": 2},
        "G": {"C": 2, "H": 3, "N": 1, "O": 1},
        "H": {"C": 6, "H": 7, "N": 3, "O": 1},
        "I": {"C": 6, "H": 11, "N": 1, "O": 1},
        "L": {"C": 6, "H": 11, "N": 1, "O": 1},
        "K": {"C": 6, "H": 12, "N": 2, "O": 1},
        "M": {"C": 5, "H": 9, "N": 1, "O": 1, "S": 1},
        "F": {"C": 9, "H": 9, "N": 1, "O": 1},
        "P": {"C": 5, "H": 7, "N": 1, "O": 1},
        "S": {"C": 3, "H": 5, "N": 1, "O": 2},
        "T": {"C": 4, "H": 7, "N": 1, "O": 2},
        "W": {"C": 11, "H": 10, "N": 2, "O": 1},
        "Y": {"C": 9, "H": 9, "N": 1, "O": 2},
        "V": {"C": 5, "H": 9, "N": 1, "O": 1},
    }

    def __init__(
        self,
        aa_seq: str | None = None,
        strict: bool = False
    ):
        if not isinstance(strict, bool):
            raise TypeError(
                "Polypeptide argument 'strict' must be bool, got "
                + type(strict).__name__
            )

        self._strict = strict
        self.aa_seq = aa_seq if aa_seq is not None else ""

    @property
    def strict(self) -> bool:
        """Getter for strict mode."""
        return self._strict

    @property
    def aa_seq(self) -> str:
        """Getter for the amino acid sequence."""
        return self._aa_seq

    @aa_seq.setter
    def aa_seq(self, aa_seq: str) -> None:
        """Setter for the amino acid sequence."""
        if aa_seq is None:
            aa_seq = ""
        if not isinstance(aa_seq, str):
            raise TypeError(
                "Polypeptide argument 'aa_seq' must be str, not "
                + type(aa_seq).__name__
            )

        aa_seq = aa_seq.upper()

        if self._strict:
            if invalid := set(aa_seq) - self.VALID_AA:
                raise ValueError(
                    f"(Strict Mode) Invalid amino acid(s) found: {invalid}"
                )

        self._aa_seq = aa_seq

    def copy(self, strict: bool | None = None) -> "Polypeptide":
        """Returns a copy of the Polypeptide instance."""
        new_strict = strict if strict is not None else self._strict
        return Polypeptide(self.aa_seq, strict=new_strict)

    def to_strict(self) -> "Polypeptide":
        """Returns a strict copy of the Polypeptide instance."""
        return self.copy(strict=True)

    def mass(self, accuracy: int = 3) -> float:
        """Calculates the molecular mass of the polypeptide."""
        if not self.aa_seq:
            return 0.0
        total = sum(self.AA_MASS.get(aa, 0) for aa in self.aa_seq) + 18.01528
        return round(total, accuracy)

    def composition(self) -> dict:
        """Returns the amino acid composition as a dictionary."""
        return dict(Counter(self.aa_seq))

    def count(self, aa: str) -> int:
        """Returns the count of a specific amino acid in the sequence."""
        return self.aa_seq.count(aa)

    def subseq(self, start: int, end: int | None = None) -> "Polypeptide":
        """Returns a subsequence from start to end (exclusive)."""
        sub = self.aa_seq[start:end]

        return Polypeptide(sub, strict=self._strict)

    def formula(self) -> str:
        """Calculates the molecular formula of the polypeptide."""
        if not self.aa_seq:
            return ""

        atom_count: defaultdict[str, int] = defaultdict(int)

        for aa in self.aa_seq:
            if aa not in self.RESIDUE_FORMULA:
                raise ValueError(
                    f"Unknown amino acid '{aa}' in sequence. "
                    "Molecular formula can only be computed for "
                    "standard amino acids."
                )

            for atom, num in self.RESIDUE_FORMULA[aa].items():
                atom_count[atom] += num

        atom_count["H"] += 2
        atom_count["O"] += 1

        order = ["C", "H"] + sorted(set(atom_count.keys()) - {"C", "H"})

        formula = []
        for element in order:
            count = atom_count[element]
            if count > 0:
                formula.append(element + (str(count) if count > 1 else ""))

        return " ".join(formula)

    def is_valid(self) -> bool:
        """Checks if the polypeptide sequence is valid."""
        return not (set(self.aa_seq) - self.VALID_AA)

    def __len__(self) -> int:
        return len(self.aa_seq)

    def __str__(self) -> str:
        return self.aa_seq

    def __repr__(self) -> str:
        from omibio.utils.truncate_repr import truncate_repr
        aa_seq_repr = truncate_repr(self.aa_seq)
        return f"Polypeptide({aa_seq_repr}, strict={self.strict})"

    def __getitem__(self, idx) -> str:
        return self.aa_seq[idx]

    def __iter__(self):
        return iter(self.aa_seq)

    def __contains__(self, item) -> bool:
        return item in self.aa_seq

    def __eq__(self, other) -> bool:
        if isinstance(other, Polypeptide):
            return self.aa_seq == other.aa_seq
        elif isinstance(other, str):
            return self.aa_seq == other.upper()

        return False

    def __add__(self, other) -> "Polypeptide":
        if isinstance(other, Polypeptide):
            strict_mode = self._strict or other.strict
            return Polypeptide(self.aa_seq + other.aa_seq, strict=strict_mode)

        elif isinstance(other, str):
            other = other.upper()
            return Polypeptide(
                self.aa_seq + other, strict=self._strict
            )

        else:
            raise TypeError(
                "Can only add Polypeptide or str to Polypeptide, got "
                + type(other).__name__
            )

    def __mul__(self, n: int) -> "Polypeptide":
        if not isinstance(n, int):
            raise TypeError(
                "Polypeptide Can only multiply by int, got "
                + type(n).__name__
            )
        if n < 0:
            raise ValueError(
                "Polypeptide cannot be multiply by a negative number"
            )

        return Polypeptide(
            self.aa_seq * n, strict=self._strict
        )


def main():
    poly1 = Polypeptide('ADQLLKAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAWS')
    print(repr(poly1))


if __name__ == "__main__":
    main()



import click
from omibio.cli.fasta_cli import fasta_group
from omibio.io import read_fasta_iter
from typing import TextIO
from omibio.sequence import Sequence

_AMBIGUOUS_BASES = {"R", "Y", "K", "M", "B", "V", "D", "H", "S", "W"}


@fasta_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
def info(source: TextIO):
    """Display information about a FASTA file."""
    result = read_fasta_iter(source)

    seq_num = 0
    total_len = 0
    longest = 0
    shortest = None

    total_gc = 0.0
    total_at = 0.0
    total_n = 0
    total_ambi = 0

    for entry in result:
        seq = entry.seq
        length = len(seq)

        seq_num += 1
        total_len += length
        longest = max(longest, length)

        if not shortest:
            shortest = length
        else:
            shortest = min(shortest, length)

        if isinstance(seq, Sequence):
            total_gc += float(seq.gc_content())
            total_at += float(seq.at_content())
        total_n += seq.count("N")
        for base in _AMBIGUOUS_BASES:
            total_ambi += seq.count(base)

    click.echo(
        f"Sequences:\t{seq_num}\n"
        f"Total length:\t{total_len} bp\n"
        f"Longest:\t{longest} bp\n"
        f"Shortest:\t{shortest} bp\n"
        f"Average length:\t{total_len // seq_num} bp\n"
        "\n"
        f"GC content:\t{(total_gc / seq_num):.3f}\n"
        f"AT content:\t{(total_at / seq_num):.3f}\n"
        f"N content:\t{(total_n / total_len):.3f} ({total_n} Ns)\n"
        f"Ambiguous:\t{(total_ambi / total_len):.3f} ({total_ambi} Ambiguous)\n"  # noqa
    )



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from omibio.sequence.sequence import Sequence


def complement(
    seq: Union["Sequence", str],
    as_str: bool = False
) -> Union["Sequence", str]:
    """Complement a given sequence.

    Args:
        seq (Sequence | str): Input sequence.
        as_str (bool, optional):
            Whether to return the result as a string. Defaults to False.

    Raises:
        TypeError: If the input sequence is not of type Sequence or string.

    Returns:
        Sequence | str: Complemented sequence.
    """
    from omibio.sequence.sequence import Sequence
    # Validate input type
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "complement() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    if isinstance(seq, str):
        seq = Sequence(seq)

    res = seq.complement()
    return (res.sequence if as_str is True else res)


def reverse_complement(
    seq: Union["Sequence", str],
    as_str: bool = False
) -> Union["Sequence", str]:
    """Reverse complement a given sequence.

    Args:
        seq (Sequence | str): Input sequence.
        as_str (bool, optional):
            Whether to return the result as a string. Defaults to False.

    Raises:
        TypeError: If the input sequence is not of type Sequence or string.

    Returns:
        str: Reverse complemented sequence.
    """
    from omibio.sequence.sequence import Sequence
    # Validate input type
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "reverse_complement() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    if isinstance(seq, str):
        seq = Sequence(seq)

    res = seq.reverse_complement()
    return str(res) if as_str is True else res


def main():
    seq = "ATGCCCTAA"
    print([complement(seq, as_str=True)])
    print([reverse_complement(seq)])


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import TYPE_CHECKING, Iterator, TextIO, cast
from os import PathLike
import warnings
if TYPE_CHECKING:
    from omibio.bio import SeqCollections, SeqEntry

VALID_NT = set("ATUCGRYKMBVDHSWN")


class FastqFormatError(Exception):

    def __init__(self, message):
        super().__init__(message)


def read_fastq_iter(
    source: str | TextIO | PathLike,
    strict: bool = False,
    warn: bool = True,
    skip_invalid_seq: bool = False
) -> Iterator["SeqEntry"]:
    """Parse a FASTA file and return a Generator.

    Raises:
        FileNotFoundError:
            If the specified file is not found.
        FastqFormatError:
            If the FASTQ format is invalid.
        FastqFormatError:
        If the sequence contains invalid characters.
        FastqFormatError:
            If the sequence name is missing.
        FastqFormatError:
            If the sequence is missing.
        FastqFormatError:
            If the quality scores length does not match the sequence length.
        FastqFormatError:
            If the '+' line is invalid.

    Yields:
        SeqEntry:
            SeqEntry objects for each sequence in the FASTQ file.
    """

    from omibio.bio import SeqEntry
    from omibio.sequence import Sequence

    if hasattr(source, "read"):
        fh = cast(TextIO, source)
        file_name = "<stdin>"
    else:
        file_path = Path(source)
        if not file_path.exists():
            raise FileNotFoundError(f"File '{source}' not found.")

        suffix = file_path.suffix.lower()
        if suffix not in {".fastq", ".fq"}:
            raise FastqFormatError(
                f"Invalid format to read: {suffix}"
            )
        file_name = str(file_path)
        fh = open(file_path, "r")

    try:
        line_num = 0

        while True:
            header = fh.readline()
            line_num += 1
            if not header:
                break

            header = header.rstrip()
            if not header:
                continue

            if not header.startswith("@"):
                raise FastqFormatError(
                    f"Line {line_num}: FASTQ header must start with '@', "
                    f"got: {header}"
                )
            seq = fh.readline()
            plus = fh.readline()
            qual = fh.readline()
            line_num += 3

            seq, plus, qual = seq.rstrip(), plus.rstrip(), qual.rstrip()

            if (not seq) or (not plus) or (not qual):
                raise FastqFormatError(
                    f"File ends prematurely at line {line_num}"
                )

            if not plus.startswith("+"):
                if strict:
                    raise FastqFormatError(
                        f"Line {line_num}: invalid '+' line: {plus}"
                    )
                elif warn:
                    warnings.warn(
                        f"Line {line_num}: invalid '+' line: {plus}, "
                        "skip record"
                    )
                continue

            if len(seq) != len(qual):
                if strict:
                    raise FastqFormatError(
                        f"Line {line_num-2} & {line_num}: Sequence / quality "
                        f"length mismatch: ({len(seq)} vs {len(qual)})"
                    )
                elif warn:
                    warnings.warn(
                        f"Line {line_num-2} & {line_num}: Sequence / quality, "
                        f"length mismatch: ({len(seq)} vs {len(qual)})"
                        "skip record"
                    )
                continue

            skip_record = False
            for char in seq.upper():
                if char not in VALID_NT:
                    if strict:
                        raise FastqFormatError(
                            f"Invalid Sequence in line {line_num-2}: {seq}"
                        )
                    elif warn:
                        warnings.warn(
                            f"Invalid Sequence in line {line_num-2}: {seq}, "
                            f"{"skip" if skip_invalid_seq else "invalid"} "
                            "record"
                        )
                    if skip_invalid_seq:
                        skip_record = True
                    break
            if skip_record:
                continue

            yield SeqEntry(
                    seq=Sequence(seq), seq_id=header[1:],
                    qual=qual, source=file_name
                )
    finally:
        if not hasattr(source, "read"):
            fh.close()


def read_fastq(
    source: str | TextIO | PathLike,
    strict: bool = False,
    warn: bool = True,
    skip_invalid_seq: bool = False
) -> "SeqCollections":
    """Parse a FASTQ file and return a SeqCollections object.

    Args:
        source (str | TextIO | PathLike):
            Path to the FASTQ file or a file-like object.
        strict (bool, optional):
            Whether to raise errors on format issues. Defaults to False.
        warn (bool, optional):
            Whether to issue warnings on format issues. Defaults to True.
        skip_invalid_seq (bool, optional):
            Whether to skip records with invalid sequences. Defaults to False.

    Returns:
        SeqCollections:
            A SeqCollections object containing the parsed sequences.
    """

    from omibio.bio import SeqCollections

    entries = []
    for entry in read_fastq_iter(
        source,
        strict=strict,
        warn=warn,
        skip_invalid_seq=skip_invalid_seq
    ):
        entries.append(entry)

    if hasattr(source, "read"):
        file_name = "<stdin>"
    else:
        file_name = str(source)

    return SeqCollections(entries=entries, source=file_name)


def main():
    input_path = r"./examples/data/example_fastq.fastq"
    result = read_fastq_iter(input_path, warn=True)
    for entry in result:
        print(entry.qual)


if __name__ == "__main__":
    main()



import click
from omibio.cli.kmer_cli import kmer_group
from omibio.io import read_fasta_iter
from omibio.analysis import kmer
from typing import TextIO
import csv


@kmer_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "-k",
    type=int,
    required=True
)
@click.option(
    "--min-count", "-min",
    type=int,
    default=1
)
@click.option(
    "--top",
    type=int,
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None,
    help="Write details to a file in csv format"
)
@click.option(
    "--canonical", "-c",
    is_flag=True
)
def count(
    source: TextIO,
    k: int,
    min_count: bool,
    canonical: bool,
    output: str | None,
    top: int | None,
):
    """Count k-mers in a FASTA file."""

    entries = read_fasta_iter(source)

    results: list[list[str | int]] = []

    for entry in entries:
        counts = kmer(
            entry.seq, k=k, canonical=canonical, min_count=min_count
        )
        for km, c in counts.items():
            results.append([entry.seq_id, k, km, c])

    tops = sorted(
        results, reverse=True, key=lambda res: res[-1]
    )

    if top is not None:
        rows = [["seq_id", "k", "kmer", "count"]] + tops[:top]
    else:
        rows = [["seq_id", "k", "kmer", "count"]] + tops
    if output is not None:
        with open(output, "w", newline="", encoding="utf-8") as f:
            writer = csv.writer(f, delimiter="\t")
            writer.writerows(rows)
            click.echo(f"Written to {output}")
    else:
        for row in rows:
            click.echo("\t".join(map(str, row)))



import click


@click.group()
def kmer_group():
    """Kmer related analysis tools."""
    pass


def register_commands():
    from .kmer_count import count
    from .kmer_total import total

    kmer_group.add_command(count)
    kmer_group.add_command(total)


register_commands()



import click


@click.version_option(
    version="0.1.3",
    prog_name="omibio",
    message="%(prog)s %(version)s"
)
@click.group()
def cli():
    """A lightweight and easy-to-use python bioinformatics toolkit."""
    pass


def register_commands():
    from omibio.cli.fasta_cli import fasta_group
    from omibio.cli.fastq_cli import fastq_group
    from omibio.cli.gc_cli import gc_group
    from omibio.cli.orf_cli import orf_group
    from omibio.cli.random_fasta_cli import random_fasta
    from omibio.cli.kmer_cli import kmer_group
    from omibio.cli.plot_cli import plot_group

    cli.add_command(gc_group, name="gc")
    cli.add_command(fasta_group, name="fasta")
    cli.add_command(fastq_group, name="fastq")
    cli.add_command(orf_group, name="orf")
    cli.add_command(kmer_group, name="kmer")
    cli.add_command(plot_group, name="plot")
    cli.add_command(random_fasta)


register_commands()



import random
from omibio.sequence import Sequence, Polypeptide


def shuffle_seq(
    seq: Sequence | Polypeptide | str,
    seed: int | None = None,
    as_str: bool = False
) -> Sequence | str | Polypeptide:
    """Return a shuffled version of the input sequence.

    Args:
        seq (Sequence | Polypeptide | str):
            The input sequence to shuffle.
        seed (int | None, optional):
            Seed for the random number generator. Defaults to None.
        as_str (bool, optional):
            Whether to return the shuffled sequence as a string.
            Defaults to False.

    Raises:
        TypeError:
            If the input types are incorrect.

    Returns:
        Sequence | str | Polypeptide:
            The shuffled sequence, either as a Sequence/Polypeptide object
            or as a string, depending on the 'as_str' parameter.
    """

    if not isinstance(seq, (Sequence, str, Polypeptide)):
        raise TypeError(
            "shuffle_seq() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )

    rng = random.Random(seed)

    chars = list(seq)
    rng.shuffle(chars)
    shuffled = "".join(chars)

    if not as_str:
        if isinstance(seq, Polypeptide):
            return Polypeptide(shuffled)
        else:
            return Sequence(shuffled)
    return shuffled


def main():
    seq = Sequence("ACGTATGATTATAGCGAGCGAGCGGGAGTTGCTGATATCTGTAC")
    print(seq.gc_content())
    shuffled = shuffle_seq(seq)
    print(shuffled)
    print(shuffled.gc_content())


if __name__ == "__main__":
    main()



from omibio.sequence.polypeptide import Polypeptide


def calc_mass(aa_seq: Polypeptide | str, accuracy: int = 3) -> float:
    """Calculate the molecular weight of a polypeptide chain molecule.

    Args:
        aa_seq (Polypeptide | str):
            A polypeptide sequence as a Polypeptide object
            or a string of amino acid single-letter codes.
        accuracy (int, optional):
            The number of decimal places to round the mass to. Defaults to 3.

    Raises:
        TypeError: If aa_seq is not a Polypeptide or str.

    Returns:
        float: The molecular weight of the polypeptide chain.
    """
    if not isinstance(aa_seq, (Polypeptide, str)):
        raise TypeError(
            "calc_mass() argument 'aa_seq' must be Polypeptide or str, got "
            + type(aa_seq).__name__
        )
    if isinstance(aa_seq, str):
        aa_seq = Polypeptide(aa_seq)

    return aa_seq.mass(accuracy)


def main():
    print(calc_mass("A"))


if __name__ == "__main__":
    main()



from omibio.sequence.sequence import Sequence


def gc(seq: Sequence | str, percent: bool = False) -> float | str:
    """Calculate the GC content of a sequence.

    Args:
        seq (Sequence | str):
            input sequence
        percent (bool, optional):
            If True, return GC content as a percentage string.
            Defaults to False.

    Returns:
        float | str: GC content as a float or percentage string.
    """
    from omibio.sequence.sequence import Sequence
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "gc() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    if isinstance(seq, Sequence):
        return seq.gc_content(percent=percent)
    elif isinstance(seq, str):
        if not seq:
            return 0.0 if not percent else "0.00%"
        gc_content = (seq.count("C") + seq.count("G")) / len(seq)
        return (
            round(gc_content, 4) if not percent
            else f"{gc_content * 100:.2f}%"
        )


def main():
    print(gc(Sequence("ACTG")))


if __name__ == "__main__":
    main()



import click


@click.group()
def gc_group():
    """GC related analysis tools."""
    pass


def register_commands():
    from .gc_compute import compute
    from .gc_sliding_window import window_gc

    gc_group.add_command(compute)
    gc_group.add_command(window_gc)


register_commands()



import click


@click.group()
def fasta_group():
    """FASTA related tools."""
    pass


def register_commands():
    from .fasta_view import view
    from .fasta_clean import clean
    from .fasta_shuffle import shuffle
    from .fasta_info import info

    fasta_group.add_command(view)
    fasta_group.add_command(clean)
    fasta_group.add_command(shuffle)
    fasta_group.add_command(info)


register_commands()



import click
from omibio.cli.fastq_cli import fastq_group
from omibio.io import read_fastq_iter
from typing import TextIO


@fastq_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "--head", "-h",
    type=int,
)
@click.option(
    "--tail", "-t",
    type=int,
)
@click.option(
    "--lengths", "-l",
    is_flag=True
)
@click.option(
    "--id-only", "-id",
    is_flag=True
)
@click.option(
    "--min-length", "-min",
    type=int,
)
@click.option(
    "--max-length", "-max",
    type=int,
)
@click.option(
    "--verbose", "-v",
    is_flag=True
)
def view(
    source: TextIO,
    head: int,
    tail: int,
    lengths: bool,
    id_only: bool,
    min_length: int,
    max_length: int,
    verbose: bool
):
    """View FASTQ file."""

    count = 0
    result = read_fastq_iter(source)

    if min_length and max_length:
        if min_length > max_length:
            raise ValueError(
                "fasta view argument 'min_length' cannot be"
                "larger than 'max_length'"
            )

    def message(entry):
        nonlocal count
        count += 1
        if id_only:
            return entry.seq_id
        elif lengths:
            return f"{entry.seq_id}\t{len(entry.seq)}"
        return f"@{entry.seq_id}\n{entry.seq}\n+\n{entry.qual}"

    if min_length or max_length:
        def check_length(entry) -> bool:
            length = len(entry.seq)
            return not (
                (min_length and length < min_length)
                or (max_length and length > max_length)
            )
    else:
        def check_length(entry) -> bool:
            return True

    if verbose:
        click.echo(f"File: {source.name}")

    if head is not None:
        for entry in result:
            if check_length(entry):
                click.echo(message(entry))
            if count >= head:
                break

    if tail is not None:
        from collections import deque
        entries: deque = deque(maxlen=tail)
        for entry in result:
            if check_length(entry):
                entries.append(entry)
        for entry in entries:
            click.echo(message(entry))

    if not head and not tail:
        for entry in result:
            if check_length(entry):
                click.echo(message(entry))

    if verbose:
        click.echo(f"All {count} sequence(s) showed")



import click
from omibio.cli.kmer_cli import kmer_group
from omibio.io import read_fasta_iter
from omibio.analysis import kmer
from collections import Counter
from typing import TextIO
import csv


@kmer_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "-k",
    type=int,
    required=True
)
@click.option(
    "--top",
    type=int,
    default=None
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None,
    help="Directory to save plots"
)
@click.option(
    "--canonical", "-c",
    is_flag=True
)
@click.option(
    "--summary", "-s",
    is_flag=True
)
def total(
    source: TextIO,
    k: int,
    canonical: bool,
    top: int | None,
    summary: bool,
    output: str | None
):
    """Count k-mers in a FASTA file."""

    entries = read_fasta_iter(source)
    total_counts: dict[str, int] = Counter()
    total = 0

    for entry in entries:
        counts = kmer(
            entry.seq, k=k, canonical=canonical
        )
        for km, c in counts.items():
            total_counts[km] += c
            total += c

    tops = sorted(
        total_counts.keys(), key=lambda n: total_counts[n], reverse=True
    )

    if summary:
        click.echo(f"k = {k}")
        click.echo(f"Total:\t{k}")
        click.echo(f"Unique:\t{len(total_counts)}\n")
        for top_kmer in tops[:top]:
            click.echo(f"{top_kmer}:\t{total_counts[top_kmer]}")
    else:
        if top is not None:
            results = [
                ["total", k, key, total_counts[key]] for key in tops[:top]
            ]
        else:
            results = [
                ["total", k, key, total_counts[key]] for key in tops
            ]
        rows = [["seq_id", "k", "kmer", "count"]] + results
        if output is not None:
            with open(output, "w", newline="", encoding="utf-8") as f:
                writer = csv.writer(f, delimiter="\t")
                writer.writerows(rows)
                click.echo(f"Written to {output}")
        else:
            for row in rows:
                click.echo("\t".join(map(str, row)))



import click
from collections import defaultdict
from omibio.cli.plot_cli import plot_group
from omibio.bio import SeqInterval
from omibio.viz import plot_sliding_gc
import matplotlib.pyplot as plt
from typing import TextIO
from math import ceil
import os
import csv


@plot_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "--output", "-o",
    type=click.Path(),
    default=None,
    help="Directory to save plots"
)
@click.option(
    "--per-page", "-p",
    type=int,
    default=3,
    help="Number of plots per page if saving images. Defaults to 3."
)
@click.option(
    "--no-show",
    is_flag=True,
    help="Whether not to show the plots."
)
def window_gc(
    source: TextIO,
    output: str,
    no_show: bool,
    per_page: int
):
    """Plot sliding window GC from a TSV file."""

    if output is not None:
        os.makedirs(output, exist_ok=True)

    figs = []

    per_page = int(per_page)
    if per_page <= 0:
        raise ValueError(
            "omibio gc sliding-window argument 'per-page' must be a "
            f"non-negative number, got {per_page}"
        )

    analysis_results = defaultdict(list)
    reader = csv.DictReader(source, delimiter="\t")
    for row in reader:
        analysis_results[row["seq_id"]].append(
            SeqInterval(
                start=int(row["start"]), end=int(row["end"]),
                seq_id=row["seq_id"], gc=float(row["gc"])
            )
        )

    seq_ids = list(analysis_results.keys())
    per_page = per_page
    page_num = ceil(len(seq_ids) / per_page)

    for page in range(page_num):
        start = page * per_page
        end = start + per_page
        page_seq_ids = seq_ids[start: end]
        n = len(page_seq_ids)
        fig, axes = plt.subplots(n, 1, figsize=(9, 3 * n))

        fig.subplots_adjust(hspace=.5)

        axes = [axes] if n == 1 else axes

        for seq_id, ax in zip(page_seq_ids, axes):
            plot_sliding_gc(
                analysis_results[seq_id], ax=ax, figsize=(6, 2)
            )

        if output is not None:
            fig.tight_layout()
            output_path = os.path.join(
                output, f"gc_page_{page+1}.png"
            )
            fig.savefig(output_path)
            if no_show:
                plt.close(fig)
            else:
                figs.append(fig)
            click.echo(f"Image saved to {output_path}")

    if not no_show:
        plt.show()

    for fig in figs:
        plt.close(fig)



import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
from matplotlib.axes import Axes
from omibio.bio.kmer_result import KmerResult


def plot_kmer(
    kmer_counts: list[KmerResult] | KmerResult,
    top_n: int = 20,
    ax=None,
    cmap="viridis",
    annot=True,
    fmt=".2f",
    show: bool = False
) -> Axes:
    """Plot a heatmap of kmer counts from KmerResult objects.

    Args:
        kmer_counts (list[KmerResult] | KmerResult):
            A list of KmerResult objects or a single KmerResult object.
        top_n (int, optional):
            Number of top kmers to display. Defaults to 20.
        ax (_type_, optional):
            Matplotlib Axes object to plot on. If None, a new figure
            and axes are created. Defaults to None.
        cmap (str, optional):
            Colormap for the heatmap. Defaults to "viridis".
        annot (bool, optional):
            Whether to annotate the heatmap cells with values.
            Defaults to True.
        fmt (str, optional):
            String format for annotations. Defaults to ".2f".
        show (bool, optional):
            Whether to display the plot immediately. Defaults to False.

    Raises:
        TypeError:
            If the input types are incorrect.
        ValueError:
            If multiple k values are found in the KmerResult objects.

    Returns:
        Axes:
            The Matplotlib Axes object containing the heatmap.
    """

    if isinstance(kmer_counts, KmerResult):
        kmer_counts = [kmer_counts]

    if not isinstance(kmer_counts, list):
        raise TypeError(
            "plot_kmer() argument 'kmer_counts' must be list[KmerResult] or "
            f"KmerResult, got {type(kmer_counts).__name__}"
        )

    kmers = []
    k_set = set()

    for count in kmer_counts:
        if not isinstance(count, KmerResult):
            raise TypeError(
                "plot_kmer() argument 'kmer_counts' must contains KmerResult, "
                f"got {type(count).__name__}"
            )
        k_set.add(count.k)
        kmers.append(pd.Series(count.counts, name=count.seq_id))

    if len(k_set) > 1:
        raise ValueError(
            f"Got multiple ks for kmers: {k_set}"
        )
    k = k_set.pop()
    df = pd.DataFrame(kmers).fillna(0)

    top_kmers = df.sum().sort_values(ascending=False).head(top_n).index
    df_top = df[top_kmers]

    df_top = df_top.div(df_top.sum(axis=1), axis=0)

    if ax is None:
        ax = plt.subplots(
            figsize=(max(8, top_n*0.5), max(6, len(kmer_counts)*0.5))
        )[1]
    sns.heatmap(df_top, annot=annot, fmt=fmt, cmap=cmap)
    ax.set_ylabel("Sequence ID")
    ax.set_xlabel(f"{k}-mers")
    ax.set_title(f"{k}-mer Heatmap")

    if show:
        plt.show()

    return ax


def main():
    from omibio.io import read_fasta_iter
    from omibio.analysis import kmer

    source = "./examples/data/example_single_short_seq.fasta"
    seq = next(read_fasta_iter(source)).seq
    res = kmer(seq, k=3, seq_id="test")
    print(repr(res))
    plot_kmer(res)
    plt.show()


if __name__ == "__main__":
    main()



import click
from omibio.cli.fasta_cli import fasta_group
from omibio.io import read_fasta, write_fasta
from typing import TextIO


@fasta_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "-o", "--output",
    type=click.Path(),
    help="Output file path."
)
@click.option(
    "--seed",
    type=int,
    default=None,
    help="Output file path."
)
def shuffle(
    source: TextIO,
    output: str | None,
    seed: int | None
):
    """
    Shuffle the sequences in the FASTA file
    and output them to the specified file.
    """
    from omibio.sequence.seq_utils.shuffle_seq import shuffle_seq
    import random

    res = {}
    rng = random.Random(seed)

    seqs = read_fasta(source).seq_dict()

    for name, seq in seqs.items():
        seq_seed = rng.randint(0, 2**32 - 1)
        shuffled = shuffle_seq(seq, seed=seq_seed, as_str=True)
        res[name] = shuffled

    if output is not None:
        write_fasta(file_name=output, seqs=res)
    else:
        for line in write_fasta(seqs=res):
            click.echo(line)



from pathlib import Path
from omibio.sequence import Sequence, Polypeptide
from omibio.bio import SeqCollections
from typing import Mapping


def write_fasta(
    seqs: Mapping[str, Sequence | Polypeptide | str] | SeqCollections,
    file_name: str | None = None,
    line_len: int = 60
) -> list[str]:
    """Write sequences to a FASTA file or return as list of strings.

    Args:
        seqs (Mapping[str, Sequence  |  Polypeptide  |  str] | SeqCollections):
            A mapping of sequence names to Sequence or Polypeptide objects,
            or a SeqCollections object.
        file_name (str | None, optional):
            Path to the output FASTA file. If None, the function
            returns the FASTA lines as a list of strings.
            Defaults to None.
        line_len (int, optional):
            Maximum line length for sequences in the FASTA output.
            Defaults to 60.

    Raises:
        TypeError:
            If the input types are incorrect.
        OSError:
            If there is an error writing to the file.

    Returns:
        list[str]:
            List of strings representing the FASTA file lines.
    """

    if not seqs:
        return []

    if isinstance(seqs, SeqCollections):
        seq_dict = seqs.seq_dict()
    elif isinstance(seqs, dict):
        seq_dict = seqs
    else:
        raise TypeError(
            "write_fasta() argument 'seqs' must be dict or SeqCollections, "
            f"got {type(seqs).__name__}"
        )

    lines = []

    for name, seq in seq_dict.items():
        if not isinstance(name, str):
            raise TypeError(
                "write_fasta() Sequence name must be str, got "
                + type(name).__name__
            )

        seq_str = str(seq).replace("\n", "")
        lines.append(f">{name}")

        for i in range(0, len(seq_str), line_len):
            lines.append(seq_str[i:i+line_len])

    if file_name is not None:
        try:
            file_path = Path(file_name)
            file_path.parent.mkdir(parents=True, exist_ok=True)
            with open(file_name, "w", encoding="utf-8") as f:
                f.writelines(line + "\n" for line in lines)

        except OSError as e:
            raise OSError(
                f"Could not write fasta to '{file_name}': {e}"
            ) from e

    return lines


def main():
    from omibio.io.read_fasta import read_fasta

    input_path = r"./examples/data/example_short_seqs.fasta"
    output_path = r"./examples/output/write_fasta_output.fasta"

    seqs = read_fasta(input_path)
    write_fasta(file_name=output_path, seqs=seqs)
    print(output_path)


if __name__ == "__main__":
    main()



from omibio.viz.plot_sliding_gc import plot_sliding_gc
from omibio.viz.plot_orfs import plot_orfs
from omibio.viz.plot_kmer_heatmap import plot_kmer
from omibio.viz.plot_motifs import plot_motifs
import matplotlib.pyplot as plt


__all__ = [
    "plt",
    "plot_sliding_gc",
    "plot_orfs",
    "plot_kmer",
    "plot_motifs"
]



from dataclasses import dataclass
from omibio.sequence.sequence import Sequence
from omibio.sequence.polypeptide import Polypeptide


@dataclass(frozen=True, slots=True)
class SeqInterval:
    """
    Stores information about the sequence range.
    """

    start: int
    end: int

    nt_seq: str | None = None

    type: str | None = None
    seq_id: str | None = None
    strand: str = "+"
    gc: float | None = None

    aa_seq: str | None = None
    frame: int = 0

    def __post_init__(self):
        if (
            self.nt_seq is not None
            and not isinstance(self.nt_seq, (str, Sequence))
        ):
            raise TypeError(
                "SeqInterval argument 'nt_seq' can only take either str "
                f"or Sequence as input, got {type(self.nt_seq).__name__}"
            )
        if isinstance(self.nt_seq, Sequence):
            object.__setattr__(self, 'nt_seq', str(self.nt_seq))

        if (
            self.aa_seq is not None
            and not isinstance(self.aa_seq, (Polypeptide, str))
        ):
            raise TypeError(
                "SeqInterval argument 'aa_seq' an only take either str"
                f"or Polypeptide as input, got {type(self.aa_seq).__name__}"
            )
        if isinstance(self.aa_seq, Polypeptide):
            object.__setattr__(self, 'aa_seq', str(self.aa_seq))

        if self.strand not in {"+", "-"}:
            raise ValueError(f"strand must be '+' or '-', got {self.strand!r}")
        if self.start < 0 or self.end < 0:
            raise ValueError("start and end must be non-negative integers")
        if self.start > self.end:
            raise ValueError(
                f"Invalid interval: start ({self.start}) > end ({self.end})"
            )
        if self.gc is not None and not isinstance(self.gc, float):
            raise TypeError(
                "SeqInterval argument 'vc' must be float, "
                f"got {type(self.gc).__name__}"
            )

    @property
    def length(self) -> int:
        """Return the length of the sequence interval."""
        return self.end - self.start

    def to_sequence(
        self, rna: bool | None = None, strict: bool = False
    ) -> Sequence:
        """Returns the nucleotide sequence as a Sequence object."""

        if self.nt_seq is not None:
            return Sequence(self.nt_seq, rna=rna, strict=strict)
        else:
            raise ValueError(
                "Cannot create Sequence: nt_seq is not set. "
            )

    def to_polypeptide(self, strict: bool = False) -> Polypeptide:
        """Returns the amino acid sequence as a Polypeptide object."""

        if self.aa_seq is not None:
            return Polypeptide(self.aa_seq, strict=strict)
        else:
            raise ValueError(
                "Cannot create Polypeptide: aa_seq is not set. "
            )

    def __len__(self) -> int:
        return self.length

    def __repr__(self) -> str:
        from omibio.utils.truncate_repr import truncate_repr
        seq_repr = truncate_repr(self.nt_seq)
        info = (
            f"SeqInterval({seq_repr}, "
            f"{self.start}-{self.end}({self.strand}), "
            f"length={self.length}"
        )
        extras = []

        if self.type is not None:
            extras.append(f"type={self.type!r}")
        if self.seq_id is not None:
            extras.append(f"seq_id={self.seq_id!r}")

        if self.aa_seq is not None:
            aa_seq_repr = truncate_repr(self.aa_seq)
            extras.append(f"aa_seq={aa_seq_repr}")
        if self.frame != 0:
            extras.append(f"frame={self.frame}")
        if self.gc is not None:
            extras.append(f"gc={self.gc!r}")

        if extras:
            info += ", " + ", ".join(extras)

        return info + ")"

    def __str__(self):
        return self.nt_seq if self.nt_seq else ""


def main():
    seq = SeqInterval(
        start=0, end=12,
        nt_seq='ATGAAAAAATAA',
        seq_id='a',
        aa_seq="MKK",
        type="ORF",
        frame=1
    )
    print(repr(seq))


if __name__ == "__main__":
    main()



from omibio.bio.seq_interval import SeqInterval
from omibio.bio.seq_entry import SeqEntry
from omibio.bio.seq_collections import SeqCollections
from omibio.bio.analysis_result import AnalysisResult
from omibio.bio.kmer_result import KmerResult
from omibio.bio.interval_result import IntervalResult

__all__ = [
    "SeqInterval",
    "AnalysisResult",
    "SeqEntry",
    "SeqCollections",
    "KmerResult",
    "IntervalResult"
]



from omibio.io.read_fasta import read_fasta, read_fasta_iter
from omibio.io.write_fasta import write_fasta
from omibio.io.read_fastq import read_fastq, read_fastq_iter
from omibio.io.write_fastq import write_fastq

__all__ = [
    "read_fasta", "read_fasta_iter",
    "write_fasta",
    "read_fastq", "read_fastq_iter",
    "write_fastq"
]



from omibio.sequence.sequence import Sequence
from omibio.bio import SeqInterval, IntervalResult
from omibio.viz.plot_motifs import plot_motifs
from typing import Pattern
import re


def find_motifs(
    seq: Sequence | str,
    pattern: str | Pattern,
    include_reverse: bool = False,
    seq_id: str | None = None,
    ignore_case: bool = True
) -> IntervalResult:
    """Finds all occurrences of a motif in a given sequence.

    Args:
        seq (Sequence | str):
            the sequence to search within.
        pattern (str | Pattern):
            the motif pattern to search for. Can be a string or compiled regex.
        seq_id (str | None, optional):
            an optional identifier for the sequence. Defaults to None.
        ignore_case (bool, optional):
            whether to ignore case when searching. Defaults to True.

    Raises:
        TypeError: If seq is not a string or Sequence.
        TypeError: If pattern is not a string or compiled Pattern.
        ValueError: If pattern is an empty string.

    Returns:
        IntervalResult:
            An object containing found motif intervals and metadata.
    """
    if not seq:
        return IntervalResult(
            intervals=[], seq_id=seq_id, type="motif", plot_func=plot_motifs,
            metadata={
                "seq_length": 0,
                "sequence": ""
            }
        )

    if not isinstance(seq, (str, Sequence)):
        raise TypeError(
            "find_motifs() argument 'seq' must be str or Sequence, "
            f"got {type(seq).__name__}"
        )

    if isinstance(pattern, str):
        if not pattern:
            raise ValueError(
                "find_motifs() argument 'pattern' cannot be an empty string"
            )
        compiled_pat = (
            re.compile(re.escape(pattern), flags=re.IGNORECASE) if ignore_case
            else re.compile(re.escape(pattern))
        )
    elif isinstance(pattern, Pattern):
        flags = pattern.flags
        if ignore_case:
            flags |= re.IGNORECASE
        else:
            flags &= ~re.IGNORECASE  # Remove IGNORECASE if present
        compiled_pat = re.compile(pattern.pattern, flags=flags)
    else:
        raise TypeError(
            "find_motifs() argument 'pattern' must be str or compiled Pattern,"
            f" got {type(pattern).__name__}"
        )

    results = []
    n = len(seq)

    def find_motifs_in_strand(seq_str: str, strand):
        for match in compiled_pat.finditer(seq_str):
            start, end = match.span()
            nt_seq = seq[start: end]
            if strand == "-":
                start, end = n - end, n - start
            results.append(
                SeqInterval(
                    start=start, end=end, nt_seq=nt_seq,
                    type='motif', seq_id=seq_id, strand=strand
                )
            )

    find_motifs_in_strand(str(seq), strand="+")
    if include_reverse:
        if isinstance(seq, str):
            seq = Sequence(seq)
        find_motifs_in_strand(str(seq.reverse_complement()), strand="-")

    return IntervalResult(
        intervals=results, seq_id=seq_id, type="motif", plot_func=plot_motifs,
        metadata={
            "seq_length": n,
            "sequence": str(seq)
        }
    )


def main():
    sequence = Sequence("AGTAATCACTGCATCGTAAGGCAGTCTTAATCGAGTCAGTC")
    res = find_motifs(sequence, "ACT", include_reverse=True)
    print(repr(res))


if __name__ == "__main__":
    main()



def truncate_repr(text: str | None, max_len: int = 30) -> str | None:
    """Truncate a string representation if it exceeds max_len.

    Args:
        text (str | None):
            The string to truncate.
        max_len (int, optional):
            Maximum length of the representation. Defaults to 30.

    Returns:
        str | None:
            Truncated representation of the string or None if text is None.
    """
    if not text:
        return None
    if len(text) <= max_len or max_len <= 3:
        return repr(text)
    half = (max_len - 3) // 2
    truncated = text[:half] + "..." + text[-half:]
    return repr(truncated)


def main():
    seq = "AGCTATGCTGATGCTAGTCTGATGCTGTAGTGCTAGTCTGTAGCACGATGCGAGTCACGATCTGATG"
    print(truncate_repr(seq))
    print(truncate_repr(None))


if __name__ == "__main__":
    main()



import click


@click.group()
def orf_group():
    """ORF related analysis tools."""
    pass


def register_commands():
    from .orf_find import find

    orf_group.add_command(find)


register_commands()



import click
from omibio.cli.fasta_cli import fasta_group
from omibio.io import read_fasta, write_fasta
from typing import TextIO, Literal


@fasta_group.command()
@click.argument(
    "source",
    type=click.File("r"),
    required=False,
    default="-"
)
@click.option(
    "-o", "--output",
    type=click.Path(),
    help="Output file path."
)
@click.option(
    "--name-policy", "-np",
    type=str,
    default="keep",
    help=(
        "Control the clean behavior of sequence names: \n"
        "'keep', 'id_only', 'underscores'"
    )
)
@click.option(
    "--gap-policy", "-gp",
    type=str,
    default="keep",
    help=(
        "Control the clean behavior of gaps: \n"
        "'keep', 'remove', 'collapse'"
    )
)
@click.option(
    "--min-len", "-min",
    type=int,
    default=10,
    help="The shortest length of the sequence to be retained"
)
@click.option(
    "--max-len", "-max",
    type=int,
    default=100_000,
    help="The longest length of the sequence to be retained"
)
@click.option(
    "--allowed-bases", "-ab",
    type=str,
    default="ATCGUNRYKMBVDHSW",
    help="Allowed bases to exist in the sequence."
)
@click.option(
    "--strict", "-s",
    is_flag=True,
    help="Whether to be strict to invalid bases."
)
@click.option(
    "--preserve-cases", "-pc",
    is_flag=True,
    help="Whether to preserve case in sequences."
)
@click.option(
    "--remove-illegal", "-ri",
    is_flag=True,
    help="Whether to remove illegal bases in non-strict mode."
)
@click.option(
    "--remove-empty", "-re",
    is_flag=True,
    help="Whether to remove sequences containing only 'N' or '-'."
)
def clean(
    source: TextIO,
    output: str | None,
    name_policy: Literal["keep", "id_only", "underscores"],
    gap_policy: Literal["keep", "remove", "collapse"],
    strict: bool,
    min_len: int,
    max_len: int,
    preserve_cases: bool,
    remove_illegal: bool,
    allowed_bases: str,
    remove_empty: bool
):
    """
    Perform data cleanup on the specified FASTA file
    and output the results to the specified file.
    """
    from omibio.sequence.seq_utils.clean import clean as c_f

    seqs = read_fasta(source, strict=False).seq_dict()
    res = c_f(
        seqs,
        name_policy=name_policy,
        gap_policy=gap_policy,
        strict=strict,
        min_len=min_len,
        max_len=max_len,
        normalize_case=not preserve_cases,
        remove_empty=remove_empty,
        remove_illegal=remove_illegal,
        allowed_bases=set(allowed_bases),
        report=False
    )
    if isinstance(res, tuple):
        cleaned = res[0]
    else:
        cleaned = res
    if output is not None:
        write_fasta(file_name=output, seqs=cleaned)
    else:
        for line in write_fasta(seqs=cleaned):
            click.echo(line)



from dataclasses import dataclass, field
from omibio.sequence import Sequence, Polypeptide
from omibio.utils.truncate_repr import truncate_repr
from typing import Any


@dataclass(slots=True)
class SeqEntry:
    """Class to hold a sequence entry with associated metadata,
    returned by file parsing functions.

    Raises:
        TypeError:
            If the input types are incorrect.
    """
    seq: Sequence | Polypeptide
    seq_id: str

    source: str | None = None
    qual: str | None = None
    metadata: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if not isinstance(self.seq, (Sequence, Polypeptide)):
            raise TypeError(
                "SeqEntry argument 'seq' must be Sequence Polypeptide, got "
                + type(self.seq).__name__
            )
        if not isinstance(self.seq_id, str):
            raise TypeError(
                "SeqEntry argument 'seq_id' must be str, got "
                + type(self.seq_id).__name__
            )

    def __str__(self) -> str:
        return str(self.seq)

    def __repr__(self) -> str:
        repr_seq = truncate_repr(str(self.seq))
        text = f"SeqEntry({repr_seq}, seq_id={self.seq_id!r}"
        if self.qual is not None:
            repr_qual = truncate_repr(self.qual)
            text += f", qual={repr_qual}"
        text += ")"
        return text


def main():
    seq = Sequence("ACTG")
    seq_entry = SeqEntry(seq, seq_id="test")
    print(repr(seq_entry))
    print(seq_entry.seq)


if __name__ == "__main__":
    main()



from pathlib import Path
from omibio.bio import SeqCollections


def write_fastq(
    seqs:  SeqCollections,
    file_name: str | None = None,
) -> list[str]:
    """Write sequences to a FASTQ file or return as list of strings.

    Args:
        seqs (SeqCollections):
            A SeqCollections object containing SeqEntry objects.
        file_name (str | None, optional):
            Path to the output FASTQ file. If None, the function
            returns the FASTQ lines as a list of strings.
            Defaults to None.

    Raises:
        TypeError:
            If the input types are incorrect.
        OSError:
            If there is an error writing to the file.

    Returns:
        list[str]:
            List of strings representing the FASTQ file lines.
    """

    if not seqs:
        return []
    if not isinstance(seqs, SeqCollections):
        raise TypeError(
            "write_fastq() argument 'seqs' must be SeqCollections, "
            f"got {type(seqs).__name__}"
        )

    lines = []

    for entry in seqs.entry_list():
        lines.append(f"@{entry.seq_id}")
        lines.append(str(entry.seq))
        lines.append("+")
        lines.append(str(entry.qual))

    if file_name is not None:
        try:
            file_path = Path(file_name)
            file_path.parent.mkdir(parents=True, exist_ok=True)
            with file_path.open("w", encoding="utf-8") as f:
                f.writelines(line + "\n" for line in lines)
        except OSError as e:
            raise OSError(
                f"Could not write fastq to '{file_name}': {e}"
            ) from e

    return lines


def main():
    from omibio.io.read_fastq import read_fastq

    input_path = r"./examples/data/example_fastq.fastq"
    output_path = r"./examples/output/write_fastq_output.fastq"

    seqs = read_fastq(input_path)
    lines = write_fastq(seqs, output_path)
    print(output_path)
    print("\n".join(lines))


if __name__ == "__main__":
    main()



from dataclasses import dataclass, field
from matplotlib.axes import Axes
from typing import Callable, Any
from abc import ABC


@dataclass
class AnalysisResult(ABC):
    """An abstract base class used to store the results of analysis functions,
    including basic information such as sequence ID, and methods for plotting.

    Args:
        ABC:
            Abstract Base Class.

    Raises:
        TypeError:
            If the input types are incorrect.
        NotImplementedError:
            If the plot method is called but no plotting function is defined.
    """

    type: str | None = None
    seq_id: str | None = None
    plot_func: Callable | None = None
    metadata: dict[str, Any] = field(default_factory=dict)

    def __post_init__(self):
        if self.type is not None and not isinstance(self.type, str):
            raise TypeError(
                "AnalysisResult argument 'type' must be str, got "
                + type(self.type).__name__
            )
        if self.seq_id is not None and not isinstance(self.seq_id, str):
            raise TypeError(
                "AnalysisResult argument 'seq_id' must be str, got "
                + type(self.seq_id).__name__
            )
        if self.plot_func is not None and not callable(self.plot_func):
            raise TypeError(
                "AnalysisResult argument 'plot_func' must be Callable, got "
                + type(self.plot_func).__name__
            )
        if not isinstance(self.metadata, dict):
            raise TypeError(
                "AnalysisResult argument 'metadata' must be dict, got "
                + type(self.plot_func).__name__
            )

    def plot(self, **kwargs) -> Axes:
        """Plot the analysis result using the associated plotting function."""
        if self.plot_func is None:
            raise NotImplementedError(
                "Plotting is not supported for this analysis result."
            )

        return self.plot_func(self, **kwargs)


def main():
    ...


if __name__ == "__main__":
    main()



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from omibio.sequence.sequence import Sequence


def transcribe(
    seq: Union["Sequence", str],
    strand: str = "+",
    as_str: bool = False
) -> Union["Sequence", str]:
    """Transcribe a DNA seq sequence to RNA.

    Args:
        seq (Sequence | str): input seq sequence
        strand (str, optional):
            Sense or antisense, either '+' or '-'. Defaults to '+'.
        as_str (bool, optional):
            Whether to return the result as a string. Defaults to False.

    Raises:
        TypeError: If the input sequence is not of type Sequence or string.

    Returns:
        Sequence | str: transcribed RNA sequence
    """

    from omibio.sequence.sequence import Sequence
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "transcribe() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    if isinstance(seq, str):
        seq = Sequence(seq)

    res = seq.transcribe(strand=strand)
    return str(res) if as_str else res


def reverse_transcribe(
    seq: Union["Sequence", str],
    as_str: bool = False
) -> Union["Sequence", str]:
    """Reverse transcribe a RNA seq sequence to DNA.

    Args:
        seq (Sequence | str): input seq sequence
        as_str (bool, optional):
        Whether to return the result as a string. Defaults to False.

    Raises:
        TypeError: If the input sequence is not of type Sequence or string.

    Returns:
        Sequence | str: reverse transcribed DNA sequence
    """
    from omibio.sequence.sequence import Sequence
    if not isinstance(seq, (Sequence, str)):
        raise TypeError(
            "reverse_transcribe() argument 'seq' must be Sequence or str, not "
            + type(seq).__name__
        )
    if isinstance(seq, str):
        seq = Sequence(seq)

    res = seq.reverse_transcribe()
    return str(res) if as_str else res


def main():
    print([transcribe("ACTG", strand="+")])
    print([reverse_transcribe("ACTG")])


if __name__ == "__main__":
    main()



from omibio.bio.seq_entry import SeqEntry
from omibio.sequence import Sequence, Polypeptide
from typing import Iterable, ItemsView, Iterator


class SeqCollections:
    """Class to hold a collection of SeqEntry objects."""

    def __init__(
        self,
        entries: Iterable[SeqEntry] | None = None,
        source: str | None = None
    ):
        """Initialization for SeqCollection.

        Args:
            entries (Iterable[SeqEntry] | None, optional):
                Iterable of SeqEntry objects to initialize the collection.
                Defaults to None.
            source (str | None, optional):
                Source information for the collection. Defaults to None.

        Raises:
            TypeError:
                If the input types are incorrect.
        """
        self._entries: dict[str, SeqEntry] = {}
        self._source = source
        if entries is not None and not isinstance(entries, Iterable):
            raise TypeError(
                "SeqCollections argument 'entries' must be Iterable "
                f"contains SeqEntry, got {type(entries).__name__}"
            )
        if entries:
            for entry in entries:
                self.add_entry(entry)

    @property
    def entries(self):
        """Return the dictionary of SeqEntry objects."""
        return self._entries

    @property
    def source(self):
        """Return the source information."""
        return self._source

    def add_entry(self, entry: SeqEntry):
        """Add a SeqEntry to the collection."""
        if not isinstance(entry, SeqEntry):
            raise TypeError(
                "SeqCollections argument 'entries' must be Iterable "
                f"contains SeqEntry, got {type(entry).__name__}"
            )
        seq_id = entry.seq_id
        if seq_id in self._entries:
            raise ValueError(
                f"Duplicate seq_id '{seq_id}'"
            )
        self._entries[seq_id] = entry

    def get_entry(self, seq_id: str) -> SeqEntry:
        """Return the SeqEntry for the given seq_id."""
        return self._entries[seq_id]

    def get_seq(self, seq_id: str) -> Sequence | Polypeptide:
        """Return the Sequence or Polypeptide for the given seq_id."""
        return self[seq_id]

    def seq_ids(self) -> list[str]:
        """Return a list of sequence IDs in the collection."""
        return list(self._entries.keys())

    def seqs(self) -> list[Sequence | Polypeptide]:
        """Return a list of sequences in the collection."""
        return [e.seq for e in self._entries.values()]

    def entry_list(self) -> list[SeqEntry]:
        """Return a list of SeqEntry objects in the collection."""
        return list(self._entries.values())

    def seq_dict(self) -> dict[str, Sequence | Polypeptide]:
        """Return a dictionary of seq_id to Sequence or Polypeptide."""
        return {e.seq_id: e.seq for e in self._entries.values()}

    def items(self) -> ItemsView[str, SeqEntry]:
        """Return an items view of the collection."""
        return self._entries.items()

    def __iter__(self) -> Iterator[SeqEntry]:
        return iter(self._entries.values())

    def __getitem__(self, seq_id: str) -> Sequence | Polypeptide:
        return self._entries[seq_id].seq

    def __contains__(self, seq_id: str) -> bool:
        return seq_id in self._entries

    def __len__(self) -> int:
        return len(self._entries)

    def __repr__(self) -> str:
        return f"SeqCollections({list(self._entries.values())!r})"

    def __str__(self) -> str:
        return str(list(self._entries.values()))


def main():
    seqs = SeqCollections(
        [
            SeqEntry(Sequence("ACTG"), seq_id="1"),
            SeqEntry(Sequence("ACTG"), seq_id="2"),
            SeqEntry(Sequence("ACTG"), seq_id="3"),
            SeqEntry(Sequence("ACTG"), seq_id="4"),
            SeqEntry(Sequence("ACTG"), seq_id="5"),
        ]
    )
    print(seqs.items())


if __name__ == "__main__":
    main()



import click


@click.group()
def plot_group():
    """Plotting related tools."""
    pass


def register_commands():
    from .plot_window_gc_cli import window_gc
    from .plot_kmer_cli import kmer

    plot_group.add_command(window_gc)
    plot_group.add_command(kmer)


register_commands()
